from __future__ import absolute_import, division, print_function

import os
import string
from abc import ABCMeta, abstractmethod
from collections import defaultdict
from numbers import Number
from typing import List, Union

import numpy as np
import seaborn as sns
import tensorflow as tf
from matplotlib import pyplot as plt
from six import add_metaclass, string_types
from tensorflow.python.keras.callbacks import Callback
from tensorflow_probability.python import distributions as tfd
from tensorflow_probability.python.distributions import Distribution

from odin.bay.distributions import ZeroInflated
from odin.visual import Visualizer
from sisua.analysis.latent_benchmarks import (plot_distance_heatmap,
                                              plot_latents_binary)
from sisua.analysis.sc_metrics import (SingleCellMetric,
                                       _preprocess_output_distribution,
                                       _to_binary)
from sisua.data import SingleCellOMIC
from sisua.data.utils import standardize_protein_name
from sisua.models import SingleCellModel
from sisua.models.base import _to_sc_omics

__all__ = ['SingleCellMonitor', 'LearningCurves', 'ScatterPlot', 'HeatmapPlot']


# ===========================================================================
# Base class
# ===========================================================================
@add_metaclass(ABCMeta)
class SingleCellMonitor(SingleCellMetric, Visualizer):

  def __init__(self, path, dpi=80, **kwargs):
    super(SingleCellMonitor, self).__init__(**kwargs)
    self._path = str(path)
    self.dpi = int(dpi)

  @property
  def path(self):
    return self._path

  def call(self, y_true: List[SingleCellOMIC], y_crpt: List[SingleCellOMIC],
           y_pred: List[Distribution], latents: List[Distribution], extras):
    self.plot(y_true, y_crpt, y_pred, latents, self.model.history.history,
              extras)
    if len(self.get_figures()) > 0:
      base, ext = os.path.splitext(self.path)
      path = base + '.%d' % (self._last_epoch + 1) + ext
      self.save_figures(path=path,
                        dpi=self.dpi,
                        separate_files=False,
                        clear_figures=True)

  @abstractmethod
  def plot(self, y_true: List[SingleCellOMIC], y_crpt: List[SingleCellOMIC],
           y_pred: List[Distribution], latents: List[Distribution], history,
           extras):
    raise NotImplementedError


class LearningCurves(SingleCellMonitor):
  """ Additional key in the loss dictionary could be given by `extras`,
  otherwise, plotting 'loss' and 'val_loss' by default.
  """

  def plot(self, y_true: List[SingleCellOMIC], y_crpt: List[SingleCellOMIC],
           y_pred: List[Distribution], latents: List[Distribution], history,
           extras):
    if extras is None:
      key = ['loss', 'val_loss']
    else:
      if isinstance(extras, string_types):
        extras = [extras]
      key = [str(i).lower() for i in extras]

    if len(key) == 0:
      return
    key = sorted([i for i in key if i in history])
    if len(key) == 0:
      return

    fig = plt.figure(figsize=(16, 4))
    for k in key:
      val = history[k]
      if k + '_epoch' in history:
        ids = history[k + '_epoch']
      else:
        ids = np.arange(len(val), dtype='int32')
      plt.plot(ids, val, label=k)
      plt.xticks(ids)
    plt.legend()
    plt.grid(True)
    self.add_figure('_'.join(key), fig)


class ScatterPlot(SingleCellMonitor):

  def __init__(self, path, data='latent', use_pca=True, dpi=80, **kwargs):
    super(ScatterPlot, self).__init__(path=path, dpi=dpi, **kwargs)
    self.use_pca = bool(use_pca)
    data = str(data.lower())
    assert data in ("latent", "imputation"), \
      "Only support three data types: latent, imputation and corrupt"
    self.data = data

  def _preprocess(self, y_true, y_pred, latents, extras):
    y_true = y_true[0]
    y_pred = [_preprocess_output_distribution(i) for i in y_pred]
    assert isinstance(extras, SingleCellOMIC), \
      "protein data must be provided as extras in form of SingleCellOMIC"
    protein = extras[y_true.indices]
    y_true.assert_matching_cells(protein)
    labels = _to_binary(protein)
    labels_name = [standardize_protein_name(i) for i in protein.var.iloc[:, 0]]

    if self.data == 'latent':
      data = latents
    elif self.data == 'imputation':
      data = y_pred
    else:
      raise NotImplementedError()
    return y_true, data, labels, labels_name

  def plot(self, y_true: List[SingleCellOMIC], y_crpt: List[SingleCellOMIC],
           y_pred: List[Distribution], latents: List[Distribution], history,
           extras):
    y_true, data, labels, labels_name = self._preprocess(
        y_true, y_pred, latents, extras)

    n = len(data)
    for idx, x in enumerate(data):
      distribution = x.__class__.__name__ \
        if not isinstance(x, tfd.Independent) else \
          x.distribution.__class__.__name__
      x = x.mean().numpy()
      if x.ndim == 3:  # in case MCMC sampling is used
        x = np.squeeze(x, axis=0)
      fig = plt.figure(figsize=(8, 8))
      plot_latents_binary(Z=x,
                          y=labels,
                          labels_name=labels_name,
                          title="[#%d]%s" % (idx, distribution),
                          use_PCA=self.use_pca,
                          ax=None,
                          show_legend=True,
                          size=12,
                          fontsize=12,
                          show_scores=False)
      self.add_figure("%s%d" % (self.data, idx), fig)


class HeatmapPlot(ScatterPlot):

  def plot(self, y_true: List[SingleCellOMIC], y_crpt: List[SingleCellOMIC],
           y_pred: List[Distribution], latents: List[Distribution], history,
           extras):
    y_true, data, labels, labels_name = self._preprocess(
        y_true, y_pred, latents, extras)

    n = len(data)
    for idx, x in enumerate(data):
      distribution = x.__class__.__name__ \
        if not isinstance(x, tfd.Independent) else \
          x.distribution.__class__.__name__
      x = x.mean().numpy()
      if x.ndim == 3:  # in case MCMC sampling is used
        x = np.squeeze(x, axis=0)
      fig = plt.figure(figsize=(8, 8))
      plot_distance_heatmap(X=x,
                            labels=labels,
                            labels_name=labels_name,
                            lognorm=True,
                            colormap='hot',
                            ax=None,
                            legend_enable=True,
                            legend_loc='upper center',
                            legend_ncol=3,
                            legend_colspace=0.2,
                            fontsize=10,
                            show_colorbar=True,
                            title="[#%d]%s" % (idx, distribution))
      self.add_figure("%s%d" % (self.data, idx), fig)



from __future__ import absolute_import, division, print_function

import inspect

import tensorflow as tf
from tensorflow.python.keras.layers import Layer

from odin.bay import Statistic, kl_divergence, print_dist
from odin.networks import DenseDistribution, Identity
from sisua.models.base import SingleCellModel
from sisua.models.latents import get_latent
from sisua.models.networks import DenseNetwork


class VariationalAutoEncoder(SingleCellModel):
  """ Variational Auto Encoder
  """

  def __init__(self,
               units,
               dispersion='full',
               xdist='zinb',
               zdist='normal',
               xdrop=0.3,
               edrop=0,
               zdrop=0,
               ddrop=0,
               hdim=128,
               zdim=32,
               nlayers=2,
               batchnorm=True,
               linear_decoder=False,
               **kwargs):
    super(VariationalAutoEncoder, self).__init__(units=units,
                                                 xdist=xdist,
                                                 dispersion=dispersion,
                                                 parameters=locals(),
                                                 **kwargs)
    self.encoder = DenseNetwork(n_units=hdim,
                                nlayers=nlayers,
                                activation='relu',
                                batchnorm=batchnorm,
                                input_dropout=xdrop,
                                output_dropout=edrop,
                                seed=self.seed,
                                name='Encoder')
    if linear_decoder:
      self.decoder = Identity(name='LinearDecoder')
    else:
      self.decoder = DenseNetwork(n_units=hdim,
                                  nlayers=nlayers,
                                  activation='relu',
                                  batchnorm=batchnorm,
                                  input_dropout=zdrop,
                                  output_dropout=ddrop,
                                  seed=self.seed,
                                  name='Decoder')
    self.latent = get_latent(zdist)(units=zdim, name=self.name + 'Latent')
    # initialize output_layers distribution
    # we must set the attribute directly so the Model will manage
    # the output layer, once all the output layers are initialized
    # the number of outputs will match `n_outputs`
    for idx, (units, posterior,
              activation) in enumerate(zip(self.units, self.xdist,
                                           self.xactiv)):
      name = 'output_layer%d' % idx
      post = DenseDistribution(
          units=units,
          posterior=posterior,
          activation=activation,
          posterior_kwargs=dict(dispersion=self.dispersion))
      setattr(self, name, post)

  def _call(self, x, lmean, lvar, t, y, masks, training, n_samples):
    # applying encoding
    e = self.encoder(x, training=training)
    # latent distribution
    qZ = self.latent(e,
                     training=training,
                     n_samples=n_samples,
                     mode=Statistic.DIST)
    # decoding the latent
    d = self.decoder(qZ.sample(n_samples), training=training)
    pX = [
        getattr(self, 'output_layer%d' % i)(d, mode=Statistic.DIST)
        for i in range(self.n_outputs)
    ]

    # note: loss is negative log-likelihood
    loss_x = self.xloss[0](t, pX[0])
    # semi-supervised NLLK
    loss_y = tf.convert_to_tensor(0, dtype=x.dtype)
    for i_true, m, i_pred, fn_loss in zip(y, masks, pX[1:], self.xloss[1:]):
      loss_y += fn_loss(i_true, i_pred) * m

    # calculating the KL
    kl = self.latent.kl_divergence(analytic_kl=self.kl_analytic,
                                   n_samples=n_samples)

    # Final ELBO
    elbo = -loss_x - loss_y - kl * self.kl_weight
    elbo = tf.reduce_logsumexp(elbo, axis=0)
    loss = tf.reduce_mean(-elbo)

    if training:
      self.add_loss(loss)
    # NOTE: add_metric should not be in control if-then-else
    self.add_metric(tf.reduce_mean(kl), aggregation='mean', name="KLqp")
    self.add_metric(tf.reduce_mean(loss_x), aggregation='mean', name="nllk_x")
    if self.is_semi_supervised:
      self.add_metric(tf.reduce_mean(loss_y), aggregation='mean', name="nllk_y")

    return pX, qZ



from __future__ import absolute_import, division, print_function

import os
import pickle
import shutil

import numpy as np

from odin import fuel as F
from odin.utils import ctext, one_hot, select_path
from sisua.data.path import DOWNLOAD_DIR, PREPROCESSED_BASE_DIR



# ===========================================================================
# Helper
# ===========================================================================
def _check_override(path, override):
  if override:
    if os.path.exists(path):
      shutil.rmtree(path)
      os.mkdir(path)
  elif not os.path.exists(path):
    os.mkdir(path)


def _preprocessing_dataset(ds, outpath):
  X_train = ds['X_train'].astype('float32').reshape(-1, 784)
  X_test = ds['X_test'].astype('float32').reshape(-1, 784)

  y_train = ds['y_train'].astype('float32')
  y_test = ds['y_test'].astype('float32')

  labels = ds['labels']
  n_classes = len(labels)

  y_train = one_hot(y_train, nb_classes=n_classes)
  y_test = one_hot(y_test, nb_classes=n_classes)

  X_train = np.concatenate((X_train, X_test), axis=0)
  y_train = np.concatenate((y_train, y_test), axis=0)
  # ====== perform normalization ====== #
  example_names = np.array(
      ["image {}".format(i + 1) for i in range(X_train.shape[0])])
  feature_names = np.array(
      ["pixel {}".format(j + 1) for j in range(X_train.shape[1])])
  # ====== saving ====== #
  data_meta = {
      'X': X_train,
      'X_row': example_names,
      'X_col': feature_names,
      'y': y_train,
      'y_col': labels
  }
  for name, val in data_meta.items():
    path = os.path.join(outpath, name)
    with open(path, 'wb') as f:
      pickle.dump(val, f)


def _validate_dataset(path):
  ds = F.Dataset(path, read_only=True)
  return ds


# ===========================================================================
# Original FMNIST
# ===========================================================================
_FMNIST_PREPROCESSED = select_path(os.path.join(PREPROCESSED_BASE_DIR,
                                                'FMNIST_preprocessed'),
                                   create_new=True)


def read_fashion_MNIST(override=False, verbose=False):
  _check_override(_FMNIST_PREPROCESSED, override)
  # ******************** load the dataset for the first time ******************** #
  if not os.path.exists(os.path.join(_FMNIST_PREPROCESSED, 'X_train')):
    ds = F.FMNIST_original.load()
    _preprocessing_dataset(ds, outpath=_FMNIST_PREPROCESSED)
  # ====== load the dataset ====== #
  return _validate_dataset(_FMNIST_PREPROCESSED)



from __future__ import print_function, division, absolute_import
import re
import numpy as np
from itertools import product

def get_arguments():
  """ Multiple values are provided by ',' """
  from odin.utils import ArgController
  args = ArgController(
  ).add('-model',
        'name for model, specified in models_.py; with the exception is "scvi" and "dca"',
        'movae'
  ).add('-ds',
        'name of the dataset: mnist, pbmc citeseq, pbmc 10xPP, cbmc, etc.',
        'pbmc8k_ly'
  ).add('-dispersion',
        'dispersion type: per "gene" or per "gene-cell"',
        'gene-cell',
        ('gene', 'gene-cell')
  ).add('-xnorm',
        'normalization method for input data: log, bin, raw, prob',
        'log',
        ('log', 'bin', 'raw', 'prob')
  ).add('-ximpu',
        'The percents for dropout entries to test imputation on X (gene count)',
        0.25,
  ).add('-tnorm',
        'normalization method for target reconstruction data',
        'raw',
        ('log', 'bin', 'raw', 'prob')
  ).add('-ynorm',
        'processing method for labels: log, bin, raw, prob',
        'prob',
        ('log', 'bin', 'raw', 'prob')
  ).add('-xclip', 'maximum value if training data to be clipped to', 0
  ).add('-yclip', 'maximum value if labels count to be clipped to', 0
  # ====== for VAE ====== #
  ).add('-xdist', 'input(X) distribution', 'zinb'
  ).add('-ydist', 'label(y) distribution', 'bernoulli'
  ).add('-zdist', 'latent distribution', 'normal'
  ).add('-cdist',
        'Distribution for the corruption method: uniform or binomial',
        'binomial'
  # ====== for training ====== #
  ).add('-lr', 'Set learning rate', 1e-4
  ).add('-epoch', 'number of training epoch', 180
  ).add('-batch', 'batch size for training', 128
  # ====== semi-supervised ====== #
  ).add('-nsample', 'number of MCMC sample during training', 1
  ).add('-ps', 'percentage of training data for supervised task', 0.8
  ).add('-hdim', 'number of dimension for hidden layer', 128
  ).add('-zdim', 'number of dimension for latent space', 32
  ).add('-nlayer', 'number of dense layers', 2
  # ====== dropout ====== #
  ).add('-xdrop', 'dropout level for input X', 0.3
  ).add('-edrop', 'dropout level for ENCODER output before the latent', 0
  ).add('-ddrop', 'dropout level for DECODER output before reconstruction', 0
  ).add('-zdrop', 'dropout level on latent space', 0
  # ====== kl-divergence ====== #
  ).add('-kl-weight',
        'weight for KL(q|p) divergence during training', 1
  ).add('-warmup',
        'Number of warmup epoch, slowly increasing KL weight to 1.', 400
  # ====== data ====== #
  ).add('-y-weight',
        'weight for supervised objective, only apply for semi-supervised system', 1
  ).add('--no-batchnorm', 'turn-off batch normalization', False
  ).add('--monitor', 'enable monitoring every epoch', False
  ).add('--no-analytic', 'turn off: analytic KL for variational inference', False
  ).add('--override', 'override previous experiments', False
  ).add('-nprocess', 'number or process to run multiple experiments at once', 1
  ).parse()
  args['batchnorm'] = not args['no_batchnorm']
  args['analytic'] = not args['no_analytic']
  n_process = int(args['nprocess'])
  del args['nprocess']
  assert n_process > 0
  # ====== post-processing ====== #
  from six import string_types
  preprocessed = {}
  for k, v in args.items():
    assert not isinstance(v, (tuple, list)), \
    "Invalid argument format, name='%s' value='%s'" % (k, str(v))
    if isinstance(v, string_types):
      v = v.split(',')
      if len(v) == 0:
        v = ''
      elif len(v) == 1:
        v = v[0]
      else:
        v = [int(i) if i.isdigit() else
            (float(i) if '.' != i and re.match('^[\+-]?\d*\.\d*$', i)
             else i)
            for i in v]
    preprocessed[k] = v
  return preprocessed, n_process

def train(model='vae', ds='pbmc8k_ly', dispersion='gene-cell',
          xnorm='log', ximpu=0.25, tnorm='raw',
          ynorm='prob',
          xclip=0, yclip=0,
          xdist='zinb', ydist='bernoulli', zdist='normal', cdist='binomial',
          lr=1e-4, epoch=128, batch=64, nsample=1, ps=0.8,
          hdim=128, zdim=32, nlayer=2,
          xdrop=0.3, edrop=0, ddrop=0, zdrop=0,
          batchnorm=True, analytic=True,
          kl_weight=1., warmup=400, y_weight=10.,
          monitor=False, override=False, **kwargs):
  import os
  os.environ['ODIN'] = 'gpu,float32,seed=5218'
  import shutil
  import pickle

  from odin.utils import (ctext, stdio)

  from sisua.data import (get_dataset, EXP_DIR)
  if model == 'scvi':
    from sisua.inference import InferenceSCVI as Inference
  elif model == 'dca':
    from sisua.inference import InferenceDCA as Inference
  else:
    from sisua.inference import Inference
  # ====== validate the arguments ====== #
  assert (0 < nlayer < 100 and
          0 < hdim < 1000 and
          0 < zdim < 1000)
  assert (0 <= ximpu < 1)
  assert (0 <= xdrop < 1 and
          0 <= edrop < 1 and
          0 <= zdrop < 1 and
          0 <= ddrop < 1)
  assert (xclip >= 0 and yclip >= 0)
  assert 0.0 < ps <= 1.0, \
      "`ps` value must be > 0 and <= 1, but given: %f" % float(ps)
  # ===========================================================================
  # Loading data
  # ===========================================================================
  DS_NAME = str(ds)
  (ds, gene_ds, prot_ds) = get_dataset(ds, override=False)
  # ====== data for training and testing ====== #
  X_train = gene_ds.get_data(data_type='train')
  y_train = prot_ds.get_data(data_type='train')
  X_test = gene_ds.get_data(data_type='test')
  y_test = prot_ds.get_data(data_type='test')
  # ===========================================================================
  # Training the network
  # ===========================================================================
  infer = Inference(
      gene_dim=X_train.shape[1],
      prot_dim=None if model in ('vae', 'scvi', 'dca')
      else y_train.shape[1],
      model=model, dispersion=dispersion,
      xnorm=xnorm, tnorm=tnorm, ynorm=ynorm,
      xclip=xclip, yclip=yclip,
      xdist=xdist, ydist=ydist, zdist=zdist,
      xdrop=xdrop, edrop=edrop, zdrop=zdrop, ddrop=ddrop,
      hdim=hdim, zdim=zdim, nlayer=nlayer,
      batchnorm=batchnorm, analytic=analytic,
      kl_weight=kl_weight, warmup=warmup, y_weight=y_weight,
      dataset=DS_NAME)
  infer.fit(X=X_train, y=y_train,
            supervised_percent=ps, validation_percent=0.1,
            corruption_rate=ximpu, corruption_dist=cdist,
            n_mcmc_samples=nsample,
            batch_size=batch, n_epoch=epoch, learning_rate=lr,
            monitoring=False, fig_dir=None,
            detail_logging=True)
  # ===========================================================================
  # Save the trained model
  # ===========================================================================
  BASE_DIR = os.path.join(EXP_DIR, ds.name)
  if not os.path.exists(BASE_DIR):
    os.mkdir(BASE_DIR)

  MODEL_DIR = os.path.join(BASE_DIR, infer.id)
  if bool(override) and os.path.exists(MODEL_DIR):
    shutil.rmtree(MODEL_DIR)
  if not os.path.exists(MODEL_DIR):
    os.mkdir(MODEL_DIR)

  # ====== print some log ====== #
  stdio(os.path.join(MODEL_DIR, 'init.log'))
  print(ctext("Model ID:", 'cyan'), ctext(infer.id, 'lightcyan'))
  print(ctext("Base dir:", 'cyan'), ctext(BASE_DIR, 'lightcyan'))
  with open(os.path.join(MODEL_DIR, 'model.pkl'), 'wb') as f:
    pickle.dump(infer, f)
  # ===========================================================================
  # Evaluation and make prediction
  # ===========================================================================
  print('\n\n')
  stdio(os.path.join(MODEL_DIR, 'score.log'))
  for name, x, y in zip(("Train", "Test"),
                        (X_train, X_test),
                        (y_train, y_test)):
    scores = infer.score(x, y, n_mcmc_samples=100)
    print("======== %s ========" % name)
    for k, v in scores.items():
      print("  ", '%-32s' % k.split(":")[0].split('/')[-1], v)
# ===========================================================================
# Calling the main
# ===========================================================================
def main():
  kw, n_process = get_arguments()
  from sisua import set_verbose
  set_verbose(True)
  multiple_args = [[(i, t) for t in j] for i, j in kw.items()
                   if isinstance(j, (tuple, list))]
  if len(multiple_args) > 0:
    from odin.utils import ctext
    from multiprocessing import Process

    multiple_args = list(product(*multiple_args))
    print("Start running %s experiments ..." %
      ctext(len(multiple_args), 'lightyellow'))

    runned_configs = {}

    processes = []
    for args in multiple_args:
      print("  \nConfigurations {%s}" %
        ', '.join(['%s:%s' % (i, ctext(j, 'cyan'))
                   for i, j in args]))
      # all new kwargs
      tmp = dict(kw)
      tmp.update(args)
      # special processing for unsupervised models
      if tmp['model'] in ('dca', 'scvi', 'vae'):
        tmp['ynorm'] = 'prob'
        tmp['yclip'] = 0
        tmp['ydist'] = 'bernoulli'
        tmp['ps'] = 0.8
        tmp['y_weight'] = 1
      # don't run duplicated experiments
      footprint = ','.join([str(i) + ':' + str(j)
                            for i, j in sorted(tmp.items())])
      if footprint in runned_configs:
        continue
      else:
        runned_configs[footprint] = 1
      # create new process for the training task
      processes.append(Process(target=train, kwargs=tmp))
      if len(processes) >= n_process:
        [p.start() for p in processes]
        [p.join() for p in processes]
        processes = []
    # don't forget to run the remains
    if len(processes) > 0:
      [p.start() for p in processes]
      [p.join() for p in processes]
  else:
    train(**kw)

if __name__ == '__main__':
  main()



from __future__ import print_function, division, absolute_import
from matplotlib import pyplot as plt
import matplotlib.mlab as mlab

import os
os.environ['ODIN'] = 'gpu,float32,seed=5218'
import pickle
import warnings
from collections import defaultdict

from tqdm import tqdm
import numpy as np
import seaborn as sns
from scipy import stats

from odin.visual import (print_dist, merge_text_graph, plot_confusion_matrix,
                         plot_figure, plot_histogram_layers, plot_save,
                         generate_random_colors, plot_histogram, Visualizer)
from odin.utils import (unique_labels, ctext, auto_logging, batching, UnitTimer,
                        ArgController, get_script_path, mpi, Progbar,
                        catch_warnings_ignore, is_number)
from odin.stats import describe

with warnings.catch_warnings():
  warnings.filterwarnings('ignore', category=ImportWarning)
  from sklearn.mixture import GaussianMixture
  from sklearn.neighbors import KernelDensity
  from sklearn.base import DensityMixin, BaseEstimator

from sisua.data.utils import standardize_protein_name

# ===========================================================================
# Helpers
# ===========================================================================
def _clipping_quartile(x, alpha=1.5, test_mode=False):
  # result wider range threshold, keep more data points,
  # lower LLK
  x = x.astype('float32')
  Q1 = np.percentile(x, q=25)
  Q3 = np.percentile(x, q=75)
  IQR = Q3 - Q1

  high = Q3 + alpha * IQR
  low = Q1 - alpha * IQR

  if test_mode:
    x[x < low] = low
    x[x > high] = high
  else:
    x = x[np.logical_and(low <= x, x <= high)]
  return x

def _log_norm(x, scale_factor=10000):
  x = x.astype('float32')
  x_sum = np.sum(x)
  return np.log1p(x / (x_sum + np.finfo(x.dtype).eps) * scale_factor)

def _draw_hist(x, ax, title, n_bins, show_yticks=True):
  count, bins = plot_histogram(
      x=x, bins=n_bins, ax=ax,
      normalize=False, kde=False, range_0_1=False,
      covariance_factor=0.25, centerlize=False, fontsize=8,
      title=title)
  plt.xlim((np.min(x), np.max(x)))
  plt.xticks(
      np.linspace(start=np.min(x), stop=np.max(x),
                  num=5, dtype='float32'),
      fontsize=6)
  if show_yticks:
    plt.yticks(
        np.linspace(start=np.min(count), stop=np.max(count),
                    num=5, dtype='int32'),
        fontsize=5)
  else:
    plt.yticks([], [])
  return count, bins

# ===========================================================================
# LabelThresholding class
# ===========================================================================
class ProbabilisticEmbedding(BaseEstimator, DensityMixin, Visualizer):

  r""" Probabilistic embedding of real values vectors using
  Gaussian-Mixture-Model

  Parameters
  ----------
  n_components_per_class : int
    number of GMM components for thresholding (default: 2)

  positive_component : int
    in case, 3 or more components are used, this is the index
    of the component from where it is positive during thresholding
    (the components are sorted by increasing mean order)

  log_norm : bool

  clip_quartile : float

  ci_threshold : float

  random_state: int

  verbose: bool

  """

  def __init__(self, n_components_per_class=2, positive_component=1,
               log_norm=True, clip_quartile=0., remove_zeros=True,
               ci_threshold=-0.68, random_state=8,
               verbose=False):
    super(ProbabilisticEmbedding, self).__init__()
    self.n_components_per_class = int(n_components_per_class)
    self.positive_component = int(positive_component)
    assert self.positive_component > 0

    self.remove_zeros = bool(remove_zeros)
    self.log_norm = bool(log_norm)
    self.clip_quartile = float(clip_quartile)

    ci_threshold = float(ci_threshold)
    assert 0 <= np.abs(ci_threshold) <= 1
    self.ci_threshold = ci_threshold

    self.verbose = bool(verbose)
    self.random_state = random_state
    self._models = []

  # ******************** properties ******************** #
  @property
  def n_classes(self):
    return len(self._models)

  @property
  def means(self):
    """ Components' means for all classes

    Returns
    -------
    array
        means array (n_components, n_classes)
    """
    return np.hstack(
        [gmm.means_.ravel()[order][:, np.newaxis]
         for order, gmm in self._models])

  @property
  def precisions(self):
    """ Components' precision for all classes

    Returns
    -------
    array
        precisions array (n_components, n_classes)
    """
    return np.hstack(
        [gmm.precisions_.ravel()[order][:, np.newaxis]
         for order, gmm in self._models])

  # ******************** main ******************** #
  def normalize(self, x, test_mode=False):
    if x.ndim > 1:
      x = x.ravel()
    assert np.all(x >= 0), "Only support non-negative values"
    if self.remove_zeros and not test_mode:
      x = x[x > 0]
    if self.clip_quartile > 0:
      x = _clipping_quartile(x, alpha=self.clip_quartile, test_mode=test_mode)
    if self.log_norm:
      x = _log_norm(x)
    return x

  def fit(self, X):
    assert X.ndim == 2, "Only support input matrix but given: %s" % str(X.shape)

    n_classes = X.shape[1]
    it = tqdm(list(range(n_classes))) if self.verbose else range(n_classes)
    for i in it:
      # ====== normalizing ====== #
      x_train = self.normalize(X[:, i], test_mode=False)
      # ====== GMM ====== #
      gmm = GaussianMixture(n_components=self.n_components_per_class,
                            covariance_type='diag',
                            init_params='kmeans',
                            n_init=8, max_iter=120,
                            random_state=self.random_state)
      gmm.fit(x_train[:, np.newaxis])
      # ====== save GMM ====== #
      means_ = gmm.means_.ravel()
      order = np.argsort(means_)
      self._models.append((order, gmm))

  def fit_transform(self, X, return_probabilities=True):
    self.fit(X)
    return self.predict_proba(X) if return_probabilities else self.predict(X)

  def _predict(self, X, threshold):
    assert X.shape[1] == self.n_classes, "Number of classes mis-match"
    y = []
    for i, (order, gmm) in enumerate(self._models):
      x_test = self.normalize(X[:, i], test_mode=True)

      # binary thresholding
      if threshold is not None:
        ci = stats.norm.interval(np.abs(threshold),
            loc=gmm.means_[order[self.positive_component]],
            scale=np.sqrt(1 / gmm.precisions_[order[self.positive_component]]))
        x_out = (x_test >= (ci[0] if threshold < 0 else ci[1])).astype('float32')
        x_out = x_out[:, np.newaxis]
      # probabilizing
      else:
        probas = gmm.predict_proba(
            x_test[:, np.newaxis]).T[order][self.positive_component:]
        probas = np.mean(probas, axis=0)
        x_out = probas[:, np.newaxis]
      y.append(x_out)
    return np.concatenate(y, axis=1)

  def predict(self, X):
    return self._predict(X, threshold=self.ci_threshold)

  def predict_proba(self, X):
    return self._predict(X, threshold=None)

  def score_samples(self, X):
    """Compute the weighted log probabilities for each sample.

    Parameters
    ----------
    X : array-like, shape (n_samples, n_features)
        List of n_features-dimensional data points. Each row
        corresponds to a single data point.

    Returns
    -------
    log_prob : array, shape (n_samples,)
        Log probabilities of each data point in X.
    """
    scores = []
    for x, (order, gmm) in zip(X.T, self._models):
      x = self.normalize(x, test_mode=True)
      s = gmm.score_samples(x[:, np.newaxis])[:, np.newaxis]
      scores.append(s)
    return np.mean(np.hstack(scores), axis=1)

  def score(self, X, y=None):
    """Compute the per-sample average log-likelihood of the given data X.

    Parameters
    ----------
    X : array-like, shape (n_samples, n_dimensions)
        List of n_features-dimensional data points. Each row
        corresponds to a single data point.

    Returns
    -------
    log_likelihood : float
        Log likelihood of the Gaussian mixture given X.
    """
    return self.score_samples(X).mean()

  # ******************** methods for diagnosing ******************** #
  def _check_input(self, X, labels):
    assert X.ndim == 2, \
    "Only support input matrix but given: %s" % str(X.shape)

    n_classes = X.shape[1]
    assert n_classes == self.n_classes, \
    "Fitted with %d classes but give %d classes" % (self.n_classes, n_classes)

    if labels is None:
      labels = ['#%d' % i for i in range(n_classes)]
    assert len(labels) == n_classes, \
    "Number of classes and labels mismatch"
    return X, labels, n_classes

  def plot_diagnosis(self, X, labels=None, n_bins=200):
    X, labels, n_classes = self._check_input(X, labels)

    nrow = n_classes
    ncol = 1
    fig = plot_figure(nrow=nrow * 2, ncol=8)
    # add 1 for threshold color
    # add 1 for PDF color
    colors = sns.color_palette(n_colors=self.n_components_per_class + 2)

    for i, (name, (order, gmm)) in enumerate(zip(labels, self._models)):
      start = ncol * i

      means_ = gmm.means_.ravel()[order]
      precision_ = gmm.precisions_.ravel()[order]
      x = self.normalize(X[:, i], test_mode=False)

      # ====== scores ====== #
      # score
      score_llk = gmm.score(x[:, np.newaxis])
      score_bic = gmm.bic(x[:, np.newaxis])
      score_aic = gmm.aic(x[:, np.newaxis])

      # ====== the histogram ====== #
      ax = plt.subplot(nrow, ncol, start + 1)
      count, bins = _draw_hist(
          x, ax=ax,
          title="[%s] LLK:%.2f BIC:%.2f AIC:%.2f" %
                (name, score_llk, score_bic, score_aic),
          n_bins=n_bins, show_yticks=True)

      # ====== draw GMM PDF ====== #
      y_ = np.exp(gmm.score_samples(bins[:, np.newaxis]))
      y_ = (y_ - np.min(y_)) / (np.max(y_) - np.min(y_)) * np.max(count)
      ax.plot(bins, y_, color='red',
              linestyle='-', linewidth=1.5, alpha=0.6)

      # ====== draw the threshold ====== #
      ci = stats.norm.interval(np.abs(self.ci_threshold),
          loc=gmm.means_[order[self.positive_component]],
          scale=np.sqrt(1 / gmm.precisions_[order[self.positive_component]]))
      threshold = ci[0] if self.ci_threshold < 0 else ci[1]
      ids = np.where(bins >= threshold, True, False)
      ax.fill_between(bins[ids], y1=0, y2=np.max(count),
                      facecolor=colors[-2], alpha=0.3)
      ax.text(np.min(bins[ids]), np.min(count), "%.2f" % threshold)

      # ====== plot GMM probability ====== #
      x_ = np.linspace(np.min(bins), np.max(bins), 1200)
      y_ = gmm.predict_proba(x_[:, np.newaxis]) * np.max(count)
      for c, j in zip(colors, y_.T):
        plt.plot(x_, j, color=c, linestyle='--', linewidth=1.8, alpha=0.6)

      # ====== draw the each Gaussian bell ====== #
      ax = ax.twinx()
      _x = np.linspace(start=np.min(x), stop=np.max(x),
                       num=800)
      for c, m, p in zip(colors, means_, precision_):
        with catch_warnings_ignore(Warning):
          j = mlab.normpdf(_x, m, np.sqrt(1 / p))
        ax.plot(_x, j, color=c, linestyle='-', linewidth=1)
        ax.scatter(_x[np.argmax(j)], np.max(j),
                   s=66, alpha=0.8, linewidth=0, color=c)
      ax.yaxis.set_ticklabels([])

    fig.tight_layout()
    self.add_figure('diagnosis', fig)
    return self

  def plot_distribution(self, X, labels=None):
    X, labels, n_classes = self._check_input(X, labels)

    X_bin = self.predict(X)
    X_prob = self.predict_proba(X)

    normalize_to_01 = lambda x: x / np.sum(x)
    dist_raw = normalize_to_01(np.sum(X, axis=0))
    dist_bin = normalize_to_01(np.sum(X_bin, axis=0))
    dist_prob = normalize_to_01(np.sum(X_prob, axis=0))
    x = np.arange(n_classes)

    fig = plot_figure(nrow=3, ncol=int(n_classes * 1.2))
    ax = plt.gca()

    colors = sns.color_palette(n_colors=3)
    bar1 = ax.bar(x, dist_raw, width=0.2, color=colors[0], alpha=0.8)
    bar2 = ax.bar(x + 0.2, dist_bin, width=0.2, color=colors[1], alpha=0.8)
    bar3 = ax.bar(x + 0.4, dist_prob, width=0.2, color=colors[2], alpha=0.8)

    ax.set_xticks(x + 0.2)
    ax.set_xticklabels(labels, rotation=-10)
    ax.legend([bar1, bar2, bar3],
              ['Original', 'Binarized', 'Probabilized'])

    ax.grid(True, axis='y')
    ax.set_axisbelow(True)
    self.add_figure('distribution', fig)
    return self

  def boxplot(self, X, labels=None):
    X, labels, n_classes = self._check_input(X, labels)

    nrow = n_classes
    ncol = 3
    fig = plot_figure(nrow=3 * nrow, ncol=int(1.5 * ncol))

    for i, (x, name) in enumerate(zip(X.T, labels)):
      start = i * ncol

      ax = plt.subplot(nrow, ncol, start + 1)
      ax.boxplot(x,
                 whis=1.5, labels=['Original'],
                 flierprops={'marker': '.', 'markersize': 8},
                 showmeans=True, meanline=True)
      ax.set_ylabel(name)

      ax = plt.subplot(nrow, ncol, start + 2)
      ax.boxplot(x[x > 0],
                 whis=1.5, labels=['NonZeros'],
                 flierprops={'marker': '.', 'markersize': 8},
                 showmeans=True, meanline=True)

      ax = plt.subplot(nrow, ncol, start + 3)
      ax.boxplot(self.normalize(x, test_mode=False),
                 whis=1.5, labels=['Normalized'],
                 flierprops={'marker': '.', 'markersize': 8},
                 showmeans=True, meanline=True)

    plt.tight_layout()
    self.add_figure('boxplot', fig)
    return self

# ===========================================================================
# Main
# ===========================================================================
def get_arguments():
  args = ArgController(
  ).add("input", "Name of the dataset or path to csv file"
  ).add("-n", "number of GMM components", 2
  ).add("-idx", "index of the positive component", 1
  ).add("-norm", "method for normalizing: raw, log", 'log', ('log', 'raw')
  ).add("-outpath", "y_bin and y_prob will be saved to this path", ''
  ).add("-figpath", "path for saving analysis figure", '/tmp/tmp.pdf'
  ).add("--verbose", "Enable verbose and saving diagnosis", False
  ).parse()
  inp = str(args.input)
  if os.path.exists(inp):
    assert os.path.isfile(inp), "%s must be path to a file" % inp
    data = []
    with open(inp, 'r') as f:
      for line in f:
        data.append(line.strip().split(','))
    data = np.array(data)
    if all(is_number(i, string_number=True) for i in data[0]):
      y_prot = data.astype('float32')
      y_prot_names = np.array(['#%d' % i for i in range(y_prot.shape[1])])
    else:
      y_prot = data[1:].astype('float32')
      y_prot_names = data[0]
    outpath = args.outpath
  else:
    from sisua.data import get_dataset
    ds, gene_ds, prot_ds = get_dataset(inp, override=False)
    y_prot = ds['y']
    y_prot_names = np.array(ds['y_col'])
    outpath = ds.path if args.outpath == '' else args.outpath
  return {'y_prot': y_prot, 'y_prot_names': y_prot_names,
          'n_components': int(args.n), 'index': int(args.idx),
          'log_norm': True if args.norm == 'log' else False,
          'outpath': outpath if len(outpath) > 0 else None,
          'figpath': args.figpath if len(args.figpath) > 0 else None,
          'verbose': bool(args.verbose)}

def main(y_prot, y_prot_names,
         n_components=2, index=1,
         log_norm=True, clip_quartile=0., remove_zeros=True,
         ci_threshold=-0.68,
         outpath=None, figpath=None,
         verbose=False):
  if outpath is not None:
    bin_path = os.path.join(outpath, 'y_bin')
    prob_path = os.path.join(outpath, 'y_prob')
  if verbose:
    print("Start label thresholding:")
    print("  Output path:", ctext(outpath, 'yellow'))
    print("  Figure path:", ctext(figpath, 'yellow'))
  # ====== protein.count ====== #
  if verbose:
    print("  Protein labels:", ctext(', '.join(y_prot_names), 'cyan'))
    print("  Protein matrix:", ctext(y_prot.shape, 'cyan'))
  # ====== already binarized ====== #
  if len(np.unique(y_prot)) == 2:
    warnings.warn("y is already binarized!")
    exit()
  # ====== PB ====== #
  pb = ProbabilisticEmbedding(
      n_components_per_class=n_components, positive_component=index,
      log_norm=log_norm, clip_quartile=clip_quartile, remove_zeros=remove_zeros,
      ci_threshold=ci_threshold, verbose=verbose)
  pb.fit(y_prot)
  y_bin = pb.predict(y_prot)
  y_prob = pb.predict_proba(y_prot)
  if verbose:
    print("  Thresholded values:")
    print("   Original     :", ctext(describe(y_prot, shorten=True), 'lightcyan'))
    print("   Binarized    :", ctext(describe(y_bin, shorten=True), 'lightcyan'))
    print("   Probabilities:", ctext(describe(y_prob, shorten=True), 'lightcyan'))
  # ====== save the results ====== #
  if outpath is not None:
    with open(bin_path, 'wb') as f:
      pickle.dump(y_bin, f)
      if verbose:
        print("  Save binarized data to:", ctext(bin_path, 'yellow'))
    with open(prob_path, 'wb') as f:
      pickle.dump(y_prob, f)
      if verbose:
        print("  Save probabilized data to:", ctext(prob_path, 'yellow'))
  # ====== save figure ====== #
  if figpath is not None:
    pb.boxplot(y_prot, y_prot_names
    ).plot_diagnosis(y_prot, y_prot_names
    ).plot_distribution(y_prot, y_prot_names
    ).save_figures(path=figpath, verbose=verbose)

if __name__ == '__main__':
  main(**get_arguments())



from __future__ import absolute_import, division, print_function

from tensorflow.python.keras import Model, Sequential
from tensorflow.python.keras.layers import (Activation, BatchNormalization,
                                            Dense, Dropout)


class DenseNetwork(Sequential):
  """ Multi-layer fully connected neural network
  """

  def __init__(self,
               n_units=128,
               nlayers=2,
               activation='relu',
               batchnorm=True,
               input_dropout=0.,
               output_dropout=0.,
               layer_dropout=0.,
               seed=8,
               name=None):
    nlayers = int(nlayers)
    layers = []
    if 0. < input_dropout < 1.:
      layers.append(Dropout(input_dropout, seed=seed))
    for i in range(nlayers):
      layers.append(
          Dense(n_units,
                activation='linear',
                use_bias=False if batchnorm else True,
                name="DenseLayer%d" % i))
      if batchnorm:
        layers.append(BatchNormalization())
      layers.append(Activation(activation))
      if layer_dropout > 0 and i != nlayers - 1:
        layers.append(Dropout(rate=layer_dropout))
    if 0. < output_dropout < 1.:
      layers.append(Dropout(output_dropout, seed=seed))
    super(DenseNetwork, self).__init__(layers=layers, name=name)


class RecurrentNetwork(Model):

  def __init__(self, name=None):
    super(RecurrentNetwork, self).__init__(name=name)
    raise NotImplementedError



from __future__ import absolute_import, division, print_function

import inspect
import os
import pickle
import time
from collections import OrderedDict, defaultdict

import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib import pyplot as plt
from six import string_types

from odin.backend import log_norm
from odin.ml import fast_pca, fast_tsne
from odin.utils import (as_tuple, cache_memory, catch_warnings_ignore, ctext,
                        flatten_list, md5_checksum)
from odin.visual import (plot_aspect, plot_confusion_matrix, plot_figure,
                         plot_frame, plot_save, plot_scatter,
                         plot_scatter_heatmap, to_axis2D)
from sisua.analysis.base import Posterior
from sisua.analysis.imputation_benchmarks import (correlation_scores,
                                                  imputation_mean_score,
                                                  imputation_score,
                                                  imputation_std_score)
from sisua.analysis.latent_benchmarks import (clustering_scores,
                                              plot_distance_heatmap,
                                              plot_latents_binary,
                                              plot_latents_multiclasses)
from sisua.data import get_dataset
from sisua.data.path import EXP_DIR
from sisua.data.utils import standardize_protein_name
from sisua.label_threshold import ProbabilisticEmbedding
from sisua.utils import filtering_experiment_path
from sisua.utils.visualization import save_figures


# ===========================================================================
# Helpers
# ===========================================================================
def _get_score_fn(score_type):
  score_type = str(score_type).lower()
  if score_type == 'imputation':
    fn_score = lambda x: x.scores_imputation
  elif score_type == 'classifier':
    fn_score = lambda x: x.scores_classifier
  elif score_type == 'cluster':
    fn_score = lambda x: x.scores_clustering
  elif score_type == 'spearman':
    fn_score = lambda x: x.scores_spearman
  elif score_type == 'pearson':
    fn_score = lambda x: x.scores_pearson
  else:
    raise ValueError("No support for score type: '%s'" % score_type)
  return fn_score

def _extract_metrics(hist):
  hist = dict(hist)
  loss = hist['loss']

  def _get(prefix, suffix, allow_empty=False):
    name = prefix + '_' + suffix
    if name in hist:
      return hist[name]
    elif 'n' + name in hist:
      return [-i for i in hist['n' + name]]
    if allow_empty:
      if prefix in hist:
        return hist[prefix]
      elif 'n' + prefix in hist:
        return [-i for i in hist['n' + prefix]]
    return None

  llkx = _get('llk', 'x', allow_empty=True)
  llky = _get('llk', 'y')
  klqpx = _get('klqp', 'x', allow_empty=True)
  klqpy = _get('klqp', 'y')
  return loss, llkx, llky, klqpx, klqpy

def _normalize_metrics(metrics):
  output = {'LLK_x': np.nan,
            'LLK_y': np.nan,
            'KLqp_x': np.nan,
            'KLqp_y': np.nan}
  #
  for name in ('llk', 'llk_x', 'nllk', 'nllk_x'):
    sign = -1 if 'n' == name[0] else 1
    if name in metrics:
      output['LLK_x'] = sign * metrics[name]
  #
  for name in ('klqp', 'klqp_x'):
    if name in metrics:
      output['KLqp_x'] = metrics[name]
  #
  for name in ('llk_y', 'nllk_y'):
    sign = -1 if 'n' == name[0] else 1
    if name in metrics:
      output['LLK_y'] = sign * metrics[name]
  #
  output['KLqp_y'] = metrics.get('klqp_y', np.nan)
  return output

# ===========================================================================
# Main
# ===========================================================================
class ResultsSheet(object):
  """ ResultsSheet """
  ALL_SCORE_TYPES = ['imputation', 'classifier', 'cluster',
                     'spearman', 'pearson']

  def __init__(self, *posteriors, verbose=True):
    super(ResultsSheet, self).__init__()
    posteriors = flatten_list(posteriors)
    assert len(posteriors) > 0, "No posteriors for analysis!"
    assert all(isinstance(i, Posterior) for i in posteriors)
    self._posteriors = posteriors
    self.verbose = bool(verbose)

  @property
  def posteriors(self):
    return self._posteriors

  def _log(self, msg):
    if self.verbose:
      print(msg)
    return self

  def save_scores(self, outpath,
                  scores=['imputation', 'classifier', 'cluster',
                          'spearman', 'pearson']):
    """ Saving score to html table format """
    ext = os.path.splitext(outpath)[-1].lower()
    if ext == '':
      outpath = outpath + '.html'
    elif ext != '.html':
      raise ValueError(
          "Only support excel with extension: html; but given '%s'" % ext)

    scores = as_tuple(scores, t=string_types)
    assert all(i in ResultsSheet.ALL_SCORE_TYPES for i in scores), \
    "Support score type %s but given %s" % (
        str(ResultsSheet.ALL_SCORE_TYPES), str(scores))
    fn_score = [_get_score_fn(s) for s in scores]

    index = []
    data = defaultdict(list)
    # store the original scores of data for
    # spearman and pearson correlation
    original = {}
    corrupted = {}

    for pos in self.posteriors:
      posid = pos.short_id.split('_')
      systype = posid[-1]
      dsname = '_'.join(posid[:2])
      posid = '_'.join(posid[2:-1])
      index.append(posid)

      pos_time = time.time()
      self._log("Calculating score for '%s' ..." % ctext(pos.short_id, 'yellow'))

      for name, fn in zip(scores, fn_score):
        start_time = time.time()

        i, j = fn(pos)
        trn = {'0_train': dsname, '1_type': systype}
        tst = {'0_test': dsname, '1_type': systype}
        trn.update(i)
        tst.update(j)
        data[name + 'train'].append(trn)
        data[name + 'test'].append(tst)

        if name in ('spearman', 'pearson'):
          trn_org, tst_org = getattr(pos, 'original_%s' % name)
          trn_org['0_train'] = dsname; tst_org['0_test'] = dsname
          trn_org['1_type'] = ''; tst_org['1_type'] = ''
          original[name + 'train'] = trn_org
          original[name + 'test'] = tst_org

          trn_crr, tst_crr = getattr(pos, 'imputed_%s' % name)
          trn_crr['0_train'] = dsname; tst_crr['0_test'] = dsname
          trn_crr['1_type'] = ''; tst_crr['1_type'] = ''
          corrupted[name + 'train'] = trn_crr
          corrupted[name + 'test'] = tst_crr

        self._log("  %s: %.2f(s)" %
          (ctext(name, 'cyan'), time.time() - start_time))
      self._log(" Finished in %.2f(s)" % (time.time() - pos_time))

      # special case of the metrics
      trn = {'0_train': dsname, '1_type': systype}
      tst = {'0_test': dsname, '1_type': systype}
      i, j = pos.scores
      trn.update(_normalize_metrics(i))
      tst.update(_normalize_metrics(j))
      data['metricstrain'].append(trn)
      data['metricstest'].append(tst)

    text = ''
    for score_type in ['metrics'] + list(scores):
      idx = list(index)
      train = list(data[score_type + 'train'])
      test = list(data[score_type + 'test'])

      if score_type in ('spearman', 'pearson'):
        train.append(corrupted[score_type + 'train'])
        train.append(original[score_type + 'train'])

        test.append(corrupted[score_type + 'test'])
        test.append(original[score_type + 'test'])

        idx.append('Corrupted')
        idx.append('Original')

      text += '<h4>Score type: "%s"<h4>' % score_type
      df = pd.DataFrame(data=train, index=idx)
      text += df.to_html(float_format='%.3f') + '\n'
      df = pd.DataFrame(data=test, index=idx)
      text += df.to_html(float_format='%.3f')

    with open(outpath, 'w') as f:
      f.write(text)
    return self._log("Saved scores at path '%s'" % ctext(outpath, 'lightyellow'))

  # ******************** comparison series ******************** #
  def plot_comparison_f1(self, test=True, model_id=lambda m: m.name,
                         fig_width=12):
    assert callable(model_id), "model_id must be callable"
    start_time = time.time()
    score_type = 'classifier'
    data_type = 'test' if bool(test) else 'train'

    n_system = len(self)
    fn_score = _get_score_fn(score_type)
    scores_name = None
    scores = []
    for pos in self.posteriors:
      s = fn_score(pos)
      s = s[1] if bool(test) else s[0]
      if score_type == 'classifier':
        del s['F1weight']
        del s['F1micro']
        del s['F1macro']
      n_labels = len(s)
      s = sorted(s.items(), key=lambda x: x[0])
      scores_name = [i[0].replace('F1_', '') for i in s]
      scores.append((model_id(pos.infer), [i[1] for i in s]))

    colors = sns.color_palette(n_colors=n_labels)
    fig, subplots = plt.subplots(nrows=1, ncols=n_system,
                                 sharey=True, squeeze=True,
                                 figsize=(int(fig_width), 2))

    for idx, (name, f1) in enumerate(scores):
      assert len(scores_name) == len(f1)
      f1_weight = np.mean(f1)
      ax = subplots[idx]
      ax.grid(True, axis='both', which='both', linewidth=0.5, alpha=0.6)
      for i, (f, c) in enumerate(zip(f1, colors)):
        ax.scatter(i, f, color=c, s=22, marker='o', alpha=0.8)
        ax.text(i - 0.2, f + 24, '%.1f' % f, fontsize=10,
                rotation=75)

      ax.plot(np.arange(n_labels), f1, linewidth=1.0, linestyle='--')
      ax.plot(np.arange(n_labels), [f1_weight for i in range(n_labels)],
              linewidth=1.2, linestyle=':', color='black',
              label=r"$\overline{F1}$:%.1f" % f1_weight)
      ax.legend(fontsize=14, loc='lower left', handletextpad=0.1, frameon=False)

      ax.set_xticks(np.arange(n_labels))
      ax.set_xlabel(name, fontsize=12)

      ax.set_ylim(-8, 130)
      ax.set_yticks(np.linspace(0, 100, 5))

      ax.xaxis.set_ticklabels([])

      ax.tick_params(axis='x', length=0)
      ax.tick_params(axis='y', length=0, labelsize=8)

      plot_frame(ax, right=False, top=False, left=True if idx == 0 else False)

    plt.tight_layout(w_pad=0)
    self.add_figure("compare_%s_%s" % (score_type, data_type), fig)

    fig = plot_figure(nrow=1, ncol=4)
    for name, c in zip(scores_name, colors):
      plt.plot(0, 0, 'o', label=name, color=c)
    plt.axis('off')
    plt.legend(ncol=int(np.ceil(len(scores_name) / 2)),
          scatterpoints=1, scatteryoffsets=[0.375, 0.5, 0.3125],
          loc='upper center', bbox_to_anchor=(0.5, -0.01),
          handletextpad=0.1, labelspacing=0., columnspacing=0.4)
    self.add_figure(
        "compare_%s_%s_legend" % (score_type, data_type), fig)

    return self._log('plot_comparison_series[%s][%s] %s(s)' %
      (score_type, data_type,
       ctext(time.time() - start_time, 'lightyellow')))
    return self

  # ******************** bar plot ******************** #
  def _bar_box_line(self,
              title, ylabel, get_score, model_id, ax, ignore=[],
              using_bar=True):
    start_time = time.time()
    assert callable(model_id), "model_id must be callable"
    assert callable(get_score)

    data = []
    for pos in self.posteriors:
      name = model_id(pos.infer)
      train, test = get_score(pos)

      for i in ignore:
        del train[i]
        del test[i]

      for i, j in train.items():
        data.append({'Model': name, ylabel: j, 'Data': 'train'})
      for i, j in test.items():
        data.append({'Model': name, ylabel: j, 'Data': 'test'})
    df = pd.DataFrame(data)

    ax = to_axis2D(ax)
    # Bar plot
    if using_bar:
      sns.barplot(x='Model', y=ylabel, hue='Data', data=df, ax=ax)
    # Box plot
    else:
      sns.boxplot(x='Model', y=ylabel, hue='Data', data=df, ax=ax)

    ax.grid(axis='y', linewidth=1.2, alpha=0.5)
    ax.set_axisbelow(True)

    self.add_figure(title, ax.get_figure())
    return self._log('%s %s(s)' %
      (title, ctext(time.time() - start_time, 'lightyellow')))

  # ******************** series plot ******************** #

  # ******************** box plot ******************** #
  def boxplot_cluster(self, model_id=lambda m: m.name, ax=None):
    return self._bar_box_line(title="boxplot_cluster", ylabel="Pooled Cluster Metrics",
                        get_score=lambda x: x.scores_clustering,
                        model_id=model_id, ax=ax, using_bar=False)

  def boxplot_f1(self, model_id=lambda m: m.name, ax=None):
    return self._bar_box_line(title="boxplot_classifier", ylabel="F1 score",
                        get_score=lambda x: x.scores_classifier,
                        model_id=model_id, ax=ax, using_bar=False,
                        ignore=['F1weight', 'F1micro', 'F1macro'])

  def boxplot_pearson(self, model_id=lambda m: m.name, ax=None):
    return self._bar_box_line(title="boxplot_pearson", ylabel="Pearson correlation",
                        get_score=lambda x: x.scores_pearson,
                        model_id=model_id, ax=ax, using_bar=False)

  def boxplot_spearman(self, model_id=lambda m: m.name, ax=None):
    return self._bar_box_line(title="boxplot_spearman", ylabel="Spearman correlation",
                        get_score=lambda x: x.scores_spearman,
                        model_id=model_id, ax=ax, using_bar=False)

  # ******************** bar plot ******************** #
  def barplot_cluster(self, model_id=lambda m: m.name, ax=None):
    return self._bar_box_line(title="barplot_cluster", ylabel="Pooled Cluster Metrics",
                        get_score=lambda x: x.scores_clustering,
                        model_id=model_id, ax=ax, using_bar=False)

  def barplot_f1(self, model_id=lambda m: m.name, ax=None):
    return self._bar_box_line(title="barplot_classifier",
                        ylabel="F1 score",
                        get_score=lambda x: x.scores_classifier,
                        model_id=model_id, ax=ax,
                        ignore=['F1weight', 'F1micro', 'F1macro'])

  def barplot_pearson(self, model_id=lambda m: m.name, ax=None):
    return self._bar_box_line(title="barplot_pearson",
                        ylabel="Pearson correlation",
                        get_score=lambda x: x.scores_pearson,
                        model_id=model_id, ax=ax)

  def barplot_spearman(self, model_id=lambda m: m.name, ax=None):
    return self._bar_box_line(title="barplot_spearman",
                        ylabel="Spearman correlation",
                        get_score=lambda x: x.scores_spearman,
                        model_id=model_id, ax=ax)

  # ******************** series plot ******************** #
  def plot_correlation_marker_pairs(self, test=True, fontsize=8):
    start_time = time.time()

    from scipy.stats import pearsonr, spearmanr
    n_system = len(self)
    data_type = 'test' if test else 'train'

    # OrderDict(name -> series)
    original_series = None
    imputed_series = []
    for pos in self:
      if test:
        v, x, y = pos.V_test, pos.X_test_org, pos.y_test
      else:
        v, x, y = pos.V_train, pos.X_train_org, pos.y_train
      if original_series is None:
        original_series = correlation_scores(
            X=x, y=y, gene_name=pos.gene_name, protein_name=pos.labels,
            return_series=True)
      imputed_series.append(
          correlation_scores(
              X=v, y=y, gene_name=pos.gene_name, protein_name=pos.labels,
          return_series=True))

    # ====== plotting ====== #
    n_pair = len(original_series)
    fig = plt.figure(figsize=(20, 5 * n_pair),
                     constrained_layout=True)
    width = 4
    grids = fig.add_gridspec(n_pair, (n_system + 1) * width)

    for row_idx, prot_gene in enumerate(original_series.keys()):
      prot_name, gene_name = prot_gene.split('/')
      original_gene, prot = original_series[prot_gene]

      # gather all series
      gene = [original_gene]
      system_name = ["Original"]
      for s, posetrior in zip(imputed_series, self.posteriors):
        i, j = s[prot_gene]
        assert np.all(prot == j)
        gene.append(i)
        system_name.append(posetrior.short_id_lines)

      # plotting each series
      for col_idx, (name, g) in enumerate(zip(system_name, gene)):
        ax = fig.add_subplot(
            grids[row_idx, width * col_idx: (width * col_idx + width)])
        ax.scatter(prot, g, s=25, alpha=0.6, linewidths=0)
        plot_aspect('auto', 'box', ax)

        title = data_type + ' - ' + prot_gene + ' - %s'if col_idx == 0 else "%s"
        title += '\nPearson:%.2f Spearman:%.2f'
        ax.set_title(title % (name,
                              pearsonr(g, prot)[0],
                              spearmanr(g, prot).correlation),
        fontsize=fontsize + (2 if col_idx == 0 else 0))
        if col_idx == 0:
          ax.set_xlabel('[Protein] %s' % prot_name, fontsize=fontsize)
          ax.set_ylabel('[Gene] %s' % gene_name, fontsize=fontsize)

        if np.mean(g) < 0.1:
          for tick in ax.yaxis.get_major_ticks():
            tick.label.set_fontsize(6)
        # ax = fig.add_subplot(
        #     grids[row_idx, (width * col_idx + width - 1): (width * col_idx + width)])
        ax = ax.twiny()
        ax.boxplot(g)
        ax.set_xticks(())
        # ax.set_xlabel(gene_name, fontsize=fontsize)

    with catch_warnings_ignore(UserWarning):
      plt.tight_layout()
    self.add_figure('correlation_%s' % data_type, fig)
    return self._log('plot_correlation_marker_pairs[%s] %s(s)' %
      (data_type, ctext(time.time() - start_time, 'lightyellow')))

  def plot_imputation_scatter(self, test=True, pca=False, color_by_library=True):
    start_time = time.time()
    n_system = len(self) + 2 # add the original and the corrupted
    data_type = 'test' if test else 'train'

    if n_system <= 5:
      nrow = 1
      ncol = n_system
    else:
      nrow = 2
      ncol = int(np.ceil(n_system / 2))

    X_org = self.posteriors[0].X_test_org if test else self.posteriors[0].X_train_org
    X_crr = self.posteriors[0].X_test if test else self.posteriors[0].X_train
    y = self.posteriors[0].y_test if test else self.posteriors[0].y_train
    labels = self.posteriors[0].labels
    is_binary_classes = self.posteriors[0].is_binary_classes
    allV = [X_org, X_crr] + [pos.V_test if test else pos.V_train
                             for pos in self.posteriors]
    assert X_org.shape == X_crr.shape and all(v.shape == X_org.shape for v in allV)
    all_names = ["[%s]Original" % data_type,
                 "[%s]Corrupted" % data_type] + [i.short_id_lines
                                                 for i in self.posteriors]

    # log-normalize everything
    if len(X_org) > 5000:
      np.random.seed(5218)
      ids = np.random.permutation(X_org.shape[0])[:5000]
      allV = [v[ids] for v in allV]
      y = y[ids]

    if is_binary_classes:
      y = np.argmax(y, axis=-1)
    else:
      y = ProbabilisticEmbedding().fit_transform(y)
      y = np.argmax(y, axis=-1)

    allV = [log_norm(v) for v in allV]

    fig = plt.figure(figsize=(min(20, 5 * ncol) + 2, nrow * 5))
    for idx, (name, v) in enumerate(zip(all_names, allV)):
      ax = plt.subplot(nrow, ncol, idx + 1)
      n = np.sum(v, axis=-1)
      v = fast_pca(v, n_components=2) if pca else fast_tsne(v, n_components=2)
      with catch_warnings_ignore(Warning):
        if color_by_library:
          plot_scatter_heatmap(x=v, val=n, ax=ax, size=8,
                               legend_enable=False,
                               grid=False, title=name)
        else:
          plot_scatter(x=v,
                       color=[labels[i] for i in y],
                       marker=[labels[i] for i in y],
                       ax=ax, size=8,
                       legend_enable=True if idx == 0 else False,
                       grid=False, title=name)

    with catch_warnings_ignore(Warning):
      plt.tight_layout()
    self.add_figure(
        'imputation_scatter_%s_%s' % ('lib' if color_by_library else 'cell',
                                      data_type), fig)
    return self._log('plot_imputation_scatter[%s] %s(s)' %
      (data_type, ctext(time.time() - start_time, 'lightyellow')))

  def plot_latents_binary_scatter(self, test=True, pca=False):
    start_time = time.time()
    data_type = 'test' if test else 'train'

    if len(self) <= 4:
      nrow = 1
      ncol = len(self)
    else:
      nrow = 2
      ncol = int(np.ceil(len(self) / 2))

    fig = plt.figure(figsize=(min(20, 5 * ncol) + 2, nrow * 5))

    for idx, pos in enumerate(self):
      ax = plt.subplot(nrow, ncol, idx + 1)
      with catch_warnings_ignore(Warning):
        pos.plot_latents_binary_scatter(test=test, ax=ax,
                                 legend=True if idx == 0 else False,
                                 pca=pca)

    with catch_warnings_ignore(Warning):
      plt.tight_layout()
    self.add_figure(
        'latents_scatter_%s' % data_type, fig)
    return self._log('plot_latents_binary_scatter[%s] %s(s)' %
      (data_type, ctext(time.time() - start_time, 'lightyellow')))

  def plot_learning_curves(self):
    start_time = time.time()

    fig = plt.figure(figsize=(20, len(self) * 4))
    n_metrics = 5

    for row_idx, pos in enumerate(self):
      row_idx = row_idx * n_metrics
      train = _extract_metrics(pos.train_history)
      valid = _extract_metrics(pos.valid_history)

      for col_idx, (name, i, j) in enumerate(
      zip(['loss', 'LLK_x', 'LLK_y', 'KLqp_x', 'KLqp_y'], train, valid)):
        col_idx += 1
        plt.subplot(len(self), n_metrics, row_idx + col_idx)

        if col_idx == 1:
          plt.title(pos.short_id_lines, fontsize=8, fontstyle='italic')
        else:
          plt.title(name)

        if i is None or j is None:
          plt.plot(0, 0); plt.xticks(()); plt.yticks(())
        else:
          plt.plot(i, linewidth=2.5, label='train:%.2f' % (np.max(i) if 'LLK' in name else np.min(i)),
                   linestyle='-')
          plt.plot(j, linewidth=2.5, label='valid:%.2f' % (np.max(j) if 'LLK' in name else np.min(j)),
                   linestyle='--', alpha=0.8)
          plt.legend()

    plt.tight_layout()
    self.add_figure('learning_curves', fig)
    return self._log('plot_learning_curves %s(s)' %
      ctext(time.time() - start_time, 'lightyellow'))

  def plot_scores(self, score_type='imputation', width=0.2):
    """
    score_type : 'imputation', 'cluster', 'spearman', 'pearson', 'classifier'
    """
    start_time = time.time()

    fn_score = _get_score_fn(score_type)
    scores = [fn_score(i) for i in self.posteriors]
    train_scores = [i[0] for i in scores]
    test_scores = [i[1] for i in scores]

    fmt = '%.4f' if score_type in ('imputation', 'cluster') \
    else '%.2f'

    n_scores = len(train_scores[0])
    n_system = len(self.posteriors)
    scores_name = sorted(list(train_scores[0].keys()))
    colors = sns.color_palette(n_colors=n_system)

    main_text_size = 10
    score_text_size = 6 if n_scores > 5 else 10

    fig = plt.figure(figsize=(18, 8))
    for idx, (name, scores) in enumerate(
        (('train', train_scores), ('test', test_scores))):
      ax = plt.subplot(2, 1, idx + 1)

      handlers = None
      start_pos = 0
      xticks = []
      for sc_idx, sc_name in enumerate(scores_name):
        sc = [i[sc_name] for i in scores]

        xids = start_pos + width * np.arange(len(sc))
        xticks.append(xids[0] + (xids[-1] - xids[0]) / 2)
        start_pos = xids[-1] + width * 2
        handlers = plt.bar(xids, sc, width=width, color=colors)
        # draw text on top of each bar
        for x, y in zip(xids, sc):
          plt.text(x, y, fmt % y,
                   fontsize=score_text_size, horizontalalignment='center')

      if idx == 0:
        plt.xticks(xticks, scores_name,
                   rotation=0, fontsize=main_text_size)
      else:
        plt.xticks(())

      plt.title('[%s]Comparison chart for %s scores' % (name, score_type),
                fontsize=main_text_size + 2)
      if idx == 1 and handlers is not None:
        ax.legend(handlers, [i.short_id for i in self.posteriors],
          loc='upper center', bbox_to_anchor=(0.5, -0.01), ncol=1,
          columnspacing=0.4, labelspacing=0.,
          fontsize=main_text_size, handletextpad=0.1)

    plt.tight_layout()
    self.add_figure('%s_compare' % score_type, fig)
    return self._log('plot_scores[%s] %s(s)' %
      (ctext(score_type, 'lightyellow'),
       ctext(time.time() - start_time, 'lightyellow')))

  # ******************** figure helper ******************** #
  def save_plots(self, path, dpi=None, separate_files=True):
    save_figures(self.figures, path, dpi, separate_files,
                clear_figures=True)
    return self

  @property
  def figures(self):
    if not hasattr(self, '_figures'):
      self._figures = OrderedDict()
    assert all(isinstance(k, string_types) and
               isinstance(v, plt.Figure)
              for k, v in self._figures.items()), \
    "Invalid stored Figures"
    return self._figures

  def add_figure(self, name, fig):
    for k, v in self.figures.items():
      if v == fig:
        return
    self.figures[name] = fig
    return self

  # ******************** properties ******************** #
  def summary(self):
    return self.__str__()

  def __str__(self):
    s = "#Posteriors: %s\n" % ctext(len(self), 'lightyellow')
    for i in self:
      s += " - %s\n" % ctext(i.id, 'cyan')
    return s

  def __len__(self):
    return len(self._posteriors)

  def __getitem__(self, key):
    if isinstance(key, string_types):
      return [i for i in self if key in i.id.split('_')][0]
    elif callable(key):
      return [i for i in self if key(i)][0]
    return self.posteriors.__getitem__(key)

  def __iter__(self):
    return self.posteriors.__iter__()



import os
import re
import math
import pickle
from six import string_types
from collections import OrderedDict, defaultdict

import numpy as np

from odin.utils import ctext, as_tuple
from odin.fuel import Dataset
from odin.stats import sparsity_percentage

def filtering_experiment_path(ds_name, incl_keywords, excl_keywords,
                              fn_filter=None,
                              return_dataset=False, print_log=False,
                              exp_path=''):
  """

  Parameters
  ----------
  ds_name : string
      direct path to experiments folder or name of the dataset

  incl_keywords : string
      list of keywords for including the experiments (connect by ',')

  excl_keywords : string
      list of keywords for excluding the experiments (connect by ',')

  exp_path : string
      optional, if not given, use SISUA_EXP

  Return
  ------
  dictionary
  corruption_config -> list of absolute path to all satisfied experiments

  Note
  ----
  only finished experiments are select, i.e. the experiment folder contain
  2 files 'config.pkl' and 'model.pkl'

  """
  from sisua.data import EXP_DIR, get_dataset
  ds_name = str(ds_name)
  if exp_path is None:
    exp_path = ''
  exp_path = str(exp_path)
  if len(exp_path) == 0:
    exp_path = EXP_DIR
  assert os.path.isdir(exp_path), exp_path
  # ====== check the keywords ====== #
  if incl_keywords is None:
    incl_keywords = []
  if excl_keywords is None:
    excl_keywords = []
  if fn_filter is None:
    fn_filter = lambda keywords: True
  # ====== get the exp path ====== #
  if ds_name is None or return_dataset:
    (ds, gene_ds, prot_ds) = get_dataset(ds_name)
    ds_name = ds.name
  exp_path = os.path.join(exp_path, ds_name)
  assert os.path.exists(exp_path), "Experiment path '%s' must exists" % exp_path
  # ====== Extract all experiments ====== #
  all_exp = []
  for name in os.listdir(exp_path):
    path = os.path.join(exp_path, name)
    # check if experiments finished
    if os.path.exists(os.path.join(path, 'model.pkl')):
      all_exp.append(path)
  all_exp = sorted(all_exp)
  # ====== start filtering ====== #
  if isinstance(incl_keywords, string_types):
    incl_keywords = [i for i in str(incl_keywords).split(',') if len(i) > 0]
  elif isinstance(incl_keywords, (tuple, list)):
    incl_keywords = as_tuple(incl_keywords, t=str)
  else:
    raise ValueError("No support for incl_keywords type: %s" % str(type(incl_keywords)))

  if isinstance(excl_keywords, string_types):
    excl_keywords = [i for i in str(excl_keywords).split(',') if len(i) > 0]
  elif isinstance(excl_keywords, (tuple, list)):
    excl_keywords = as_tuple(excl_keywords, t=str)
  else:
    raise ValueError("No support for excl_keywords type: %s" % str(type(excl_keywords)))

  all_exp = [i for i in all_exp
             if all(any(j in keyword
                        for keyword in os.path.basename(i).split('_'))
                    for j in incl_keywords)]
  all_exp = [i for i in all_exp
             if all(all(j not in keyword
                        for keyword in os.path.basename(i).split('_'))
                    for j in excl_keywords)]

  # filter function
  all_exp = [i for i in all_exp
             if fn_filter(os.path.basename(i).split('_'))]
  # ====== logging ====== #
  if bool(print_log):
    print(ctext("Found following experiments:", 'lightyellow'))
    for name, paths in all_exp.items():
      print("*", ctext(name, 'yellow'))
      for i in paths:
        print('  ', os.path.basename(i))

  if return_dataset:
    return all_exp, ds, gene_ds, prot_ds
  return all_exp

def anything2image(x):
  if x.ndim == 1:
    img_dim = int(math.ceil(math.sqrt(x.shape[0])))
    z = np.zeros(shape=img_dim * img_dim, dtype=x.dtype)
    z[:x.shape[0]] = x
    x = z.reshape(img_dim, img_dim)
  elif x.ndim == 2:
    pass
  elif x.ndim == 3:
    pass
  else:
    raise ValueError("No support for image with %d dimensions" % x.ndim)
  return x

# ===========================================================================
# For thresholding based on validation dataset
# ===========================================================================
def apply_threshold(x, threshold):
  """
  x = 0 if x < threshold
  elif threshold <= x < 1, x = 1
  otherwise, x
  """
  x = np.where(x < threshold, 0, x)
  x = np.where(np.logical_and(0 < x, x < 1), 1, x).astype('int32')
  return x

def thresholding_by_sparsity_matching(T, W, *applying_data):
  """
  T : original count
  W : reconstructed count
  """
  T = T.astype('int32')
  W = W[0] if isinstance(W, (tuple, list)) else W
  if W.ndim == 3:
    W = W[0]
  assert W.ndim == 2

  best_threshold = None
  nonzeros = T != 0
  # this is just matching the sparsity
  for threshold in np.linspace(0, 1, num=100, endpoint=True)[::-1]:
    tmp = W >= threshold
    if np.sum(tmp) >= np.sum(nonzeros):
      best_threshold = threshold
      break
  # ====== applying the threshold ====== #
  new_data = []
  for data in applying_data:
    if data is None:
      new_data.append(None)
    else:
      if isinstance(data, tuple):
        data = list(data)
      if isinstance(data, list) or data.ndim == 3:
        data[0] = apply_threshold(data[0], threshold=best_threshold)
      else:
        data = apply_threshold(data, threshold=best_threshold)
      new_data.append(data)
  return best_threshold, tuple(new_data)



from __future__ import print_function, division, absolute_import
import os
import re
import io
import copy
import types
import pickle
import inspect
import warnings
from itertools import chain
from six import string_types
from functools import partial
from collections import OrderedDict, defaultdict

import numpy as np
from sklearn.base import BaseEstimator

import tensorflow as tf
from tensorflow.python.keras import Input
from tensorflow.python.keras.callbacks import LambdaCallback
from tensorflow.python.keras.layers import (Lambda as KerasLambda,
                                            Add as KerasAdd,
                                            Concatenate as KerasConcatenate)
from tensorflow.python.util.nest import flatten


from odin.fuel import Dataset
from odin.ml import fast_pca, fast_tsne
from odin.stats import train_valid_test_split, describe
from odin.utils import (get_script_path, ctext, batching, one_hot, uuid,
                        as_list, get_module_from_path, stdio,
                        get_formatted_datetime)
from odin import nnet as N, backend as K, training
from odin.backend.keras_helpers import (has_keras_meta, to_keras_objective,
                                        copy_keras_metadata)

import sisua
from sisua import is_verbose
from sisua.data import get_library_size, apply_artificial_corruption
from sisua.data.const import UNIVERSAL_RANDOM_SEED
from sisua.data import get_dataset, SingleCellOMIC
from sisua.utils import LearningCurves, plot_monitoring_epoch
from sisua.label_threshold import ProbabilisticEmbedding

from ._consts import PREDICTION_BATCH_SIZE

# The following tensors are allowed to return
possible_outputs = (
    'Z', 'Z_sample', # latent space
    'loss', # a tensor for training
    'metr', # metrics for monitoring
    # library size
    'L', 'L_sample', 'L_stddev',
    # W is predicted corrupted count (with dropout)
    'W', 'W_sample', 'W_stddev',
    # V is the cleaned count (without dropout)
    'V', 'V_sample', 'V_stddev',
    'y', 'y_sample', # predict values of protein markers
    'pi' # the zero inflated rate
)

def _signature_error(args):
  print("Function must contain following arguments:")
  print('  X      : [n_sample, n_gene]    input gene expression')
  print('  T      : [n_sample, n_gene]    target for the reconstruction')
  print('  L      : [n_sample, 1]         library size')
  print('  L_mean : [n_sample, 1]         mean of library size')
  print('  L_var  : [n_sample, 1]         variance of library')
  print('  mask   : [n_sample, 1]         mask for supervised training')
  print('  y      : [n_sample, n_protein] protein markers as labels')
  print('  nsample: []                    a scalar, number of MCMC sample')
  print('  nepoch : []                    a scalar, number of trained epoch')
  print('  configs: dict                  dictionary for the configuration')
  raise RuntimeError("But given: %s" % args)

def _normalize_tensor_name(name):
  name = name.split(':')[0].split('/')[-1].lower()
  name = re.sub(r"_\d+$", '', name) # e.g. 'loss_1'
  return name

# ===========================================================================
# Inference class
# ===========================================================================
class Inference(BaseEstimator):
  """ Performing inference given pre-defined model

  Notation
  --------
  X : [n_samples, n_genes]
      the gene expression matrix for single cell data, each row is
      one cell

  T : [n_samples, n_genes]
      the target variables for the reconstruction objective, each row
      is a corresponding cell data for `X`

  C : [n_samples, 1]
      the size factor (i.e. count-sum, library size) for each cell
      in each row

  y : [n_samples, n_proteins]
      the protein markers level for each single cell, each row is
      one cell

  Z : [n_samples, zdim]
      the latent space, mean of the latent distribution

  Parameters
  ----------
  model_name : string
      name of the model defined in `sisua.models`

  dispersion : {'gene', 'gene-cell'}
      'gene' - dispersion parameter is constant per gene across cells
      'gene-cell' - dispersion parameter can differ for every gene in every cell

  hdim : number of hidden units for each hidden layer

  zdim : number of latent dimension

  nlayer : number of layers for both encoder and decoder

  xnorm : {'raw', 'log', 'bin', 'prob'}
  tnorm : {'raw', 'log', 'bin', 'prob'}
  ynorm : {'raw', 'log', 'bin', 'prob'}
      different strategies for data normalization
      raw - keep the raw data (i.e. count or expression level)
      log - log-normalization
      bin - using `sisua.label_threshold.ProbabilisticEmbedding` to convert
            data to binary format
      prob - using `sisua.label_threshold.ProbabilisticEmbedding` to convert
             data to probability values

  xdist : the distribution name for X
  ydist : the distribution name for y
      support distribution include:
      'bernoulli' - Bernoulli distribution for binary data
      'zibernoulli' - Zero-inflated Bernoulli distribution for binary data
      'normal' - normal (or Gaussian) distribution
      'nb' - negative binomial for raw count data
      'zinb' - Zero-inflated negative binomial
      'poisson' - Poisson distribution for rare count event
      'zipoisson' - Zero-inflated Poisson
      'beta' - Beta distribution

  zdist : the distribution name for Z

  batchnorm : enable batch-normalization

  xdrop : dropout level for input X (from 0. to 1.)
  edrop : dropout level for encoder output (from 0. to 1.)
  zdrop : dropout level for latent space (from 0. to 1.)
  ddrop : dropout level for decoder output (from 0. to 1.)
      Dropout is enable if the value is greater than 0

  analytic : bool (default: True)
      using analytic KL-divergence

  kl_weight : scalar
      fixed weight for KL-divergence on latent

  warmup : integer (> 0)
      number of warmup epoch for KL-divergence

  extra_module_path : {string, None}
      path to folder contain the model script that you defined
      yourself, the .py file should has 'models_' prefix
      (e.g. 'models_vae.py')

  """

  def __init__(self, gene_dim, prot_dim=None,
               model='vae', dispersion='gene-cell',
               xnorm='log', tnorm='raw', ynorm='prob',
               xclip=0, yclip=0,
               xdist='zinb', ydist='bernoulli', zdist='normal',
               xdrop=0.3, edrop=0, zdrop=0, ddrop=0,
               hdim=128, zdim=32, nlayer=2,
               batchnorm=True, analytic=True,
               kl_weight=1., warmup=400, y_weight=10.,
               extra_module_path=None,
               **kwargs):
    super(Inference, self).__init__()
    assert int(warmup) > 0, "Number of warmup epoch must > 0"
    # ====== dispersion value ====== #
    assert dispersion in ('gene', 'gene-cell'), \
    "Only support 'gene' or 'gene-cell' dispersion"
    configs = dict(locals())
    # ====== modify the config ====== #
    primary_path = [os.path.join(os.path.dirname(inspect.getfile(sisua)), 'models')]
    extra_module_path = ([] if extra_module_path is None else
                         as_list(extra_module_path, t=string_types))
    configs['module_name'] = str(model)
    configs['module_path'] = primary_path + extra_module_path
    # remove unnecessary
    del configs['model']
    del configs['extra_module_path']
    del configs['self']
    del configs['kwargs']
    del configs['__class__']
    configs.update(kwargs)
    # select objective for multitask model
    # ce : categorical cross entropy
    # mse : mean squared error
    # ll : log loss
    # sg : multiple sigmoid loss
    configs['rec_loss'] = 'mse'
    if ynorm == 'raw' or ynorm == 'log':
      configs['cls_loss'] = 'mse'
    else:
      configs['cls_loss'] = 'sg'
    # ====== given the model name, searching for it ====== #
    if isinstance(model, string_types):
      all_error_while_loading = {}
      found = []
      for p in primary_path + extra_module_path:
        m, e = get_module_from_path(
            identifier=model,
            path=p,
            prefix='models_',
            return_error=True)
        all_error_while_loading[p] = e
        found += m
      if len(found) == 0:
        print("The following errors happened while loading script:")
        for i, j in all_error_while_loading.items():
          print('Path:', i)
          print(str(j))
        raise RuntimeError("No model found!")
      elif len(found) > 1:
        raise RuntimeError("Duplicated model found: %s" % str(model))
      self._model = found[0]
      if isinstance(self._model, N.NNOp):
        configs['is_keras_model'] = False
      else:
        self._model = self._model()
        configs['is_keras_model'] = True
    # error no support
    else:
      raise ValueError("No support for `model` with type: %s" % str(type(model)))

    # ====== basics ====== #
    self._training_log = ''
    self._history = {'train': defaultdict(list),
                     'valid': defaultdict(list)}
    self._configs = configs

    self._gmm_threshold = {}
    self._gene_dim = int(gene_dim)
    self._prot_dim = 0 if prot_dim is None else max(int(prot_dim), 0)
    self._is_fitted = False

    self._ps = 0
    self._trained_n_epoch = 0
    self._batch_size = 0
    self._learning_rate = 0
    self._n_mcmc_train = 0
    self._corruption_rate = None
    self._corruption_dist = None
    # ====== validate the model ====== #
    if self.is_keras_model:
      self._name = self._model.__class__.__name__
    else:
      self._name = self._model.name
    configs['model_name'] = self._name
    if hasattr(self._model, '__call__'):
      # X, T, L, L_mean, L_var, mask, y, nsample, nepoch, kwargs
      sign = inspect.getargspec(self._model.__call__
                                if self.is_keras_model else
                                self._model._func._func)
      func_args = sign.args
      if 'self' == func_args[0]:
        func_args = func_args[1:]
      require_args = ['X', 'T',
                      'L', 'L_mean', 'L_var',
                      'mask', 'y',
                      'nsample', 'nepoch', 'configs']
      assert func_args == require_args, _signature_error(sign.args)
    else:
      raise ValueError(
          "Design model must be callable! Found object: %s" % self._model)
    # ====== init ====== #
    self._reset()
    self._initialize_placeholders_and_outputs()

  def _reset(self):
    self.input_plhs = None
    self.n_mcmc_samples_var = None
    self.n_epochs_var = None
    self._outputs = defaultdict(dict)
    self._other_outputs = defaultdict(dict)
    self._pred_functions = {}
    self.score_functions = {}

  # ******************** pickling ******************** #
  def get_states(self):
    # clean the configs
    for k, v in list(self._configs.items()):
      if isinstance(v, types.ModuleType):
        del self._configs[k]
    if 'self' in self._configs:
      del self._configs['self']
    if 'kwargs' in self._configs:
      del self._configs['kwargs']
    states = (self._name,
              self._configs, self._history, self._training_log,
              self._gmm_threshold, self._gene_dim, self._prot_dim,
              self._is_fitted,
              self._ps,
              self._trained_n_epoch,
              self._batch_size,
              self._learning_rate,
              self._n_mcmc_train,
              self._corruption_rate, self._corruption_dist)
    return states

  def set_states(self, states):
    (self._name,
     self._configs, self._history, self._training_log,
     self._gmm_threshold, self._gene_dim, self._prot_dim,
     self._is_fitted,
     self._ps,
     self._trained_n_epoch,
     self._batch_size,
     self._learning_rate,
     self._n_mcmc_train,
     self._corruption_rate, self._corruption_dist) = states

  def __getstate__(self):
    if not self.is_keras_model:
      model = N.serialize(nnops=self._model, path=None,
                          save_variables=True,
                          binary_output=True)
    else:
      model = self._model.get_weights()
    return self.get_states(), model

  def __setstate__(self, states):
    states, model = states
    # set the states
    self.set_states(states)
    # reload the model
    if not self.is_keras_model:
      self._model = N.deserialize(model, force_restore_vars=True)
    else: # for keras, we have to search for the function again
      found = []
      for p in self.configs['module_path']:
        m, e = get_module_from_path(
            identifier=self.configs['module_name'],
            path=p,
            prefix='models_',
            return_error=True)
        found += m
      self._model = found[0](model)
    # make sure everything initialized
    self._reset()
    self._initialize_placeholders_and_outputs()

  # ******************** initialization ******************** #
  def _initialize_placeholders_and_outputs(self):
    if self.input_plhs is not None and \
    len(self._outputs) > 0:
      return

    with tf.variable_scope(self.name + '_placeholders'):
      X_plh = Input(shape=(self.gene_dim,), dtype='float32', name='Gene_input')
      T_plh = Input(shape=(self.gene_dim,), dtype='float32', name="Gene_target")
      # mask for supervised training
      if self.prot_dim > 0:
        mask_plh = Input(shape=(), dtype='float32', name='mask')
        y_plh = Input(shape=(self.prot_dim,), dtype='float32', name='Protein')
      else:
        mask_plh = None
        y_plh = None
      # size factor
      L_plh = Input(shape=(1,), dtype='float32', name='CellSize')
      L_mean_plh = Input(shape=(1,), dtype='float32', name="CellSize_mean")
      L_var_plh = Input(shape=(1,), dtype='float32', name="CellSize_variance")
      # number of sample for MCMC
      self.n_mcmc_samples_var = K.variable(1, dtype='int32',
                                           name='n_mcmc_samples',
                                           initialize=True)
      self.n_epochs_var = K.variable(self._trained_n_epoch, dtype='float32',
                                     name='n_epochs',
                                     initialize=True)
    # all input in specific order
    self.input_plhs = (X_plh, T_plh,
                       L_plh, L_mean_plh, L_var_plh,
                       mask_plh, y_plh)
    if is_verbose():
      print(ctext("Input placeholders:", 'lightyellow'))
      for i in self.input_plhs:
        print(" *", ctext(i, 'cyan'))
    # ====== Keras ====== #
    input_dicts = {
        'X': X_plh, 'T': T_plh,
        'L': L_plh, 'L_mean': L_mean_plh, 'L_var': L_var_plh,
        'mask': mask_plh, 'y': y_plh,
        'nsample': self.n_mcmc_samples_var,
        'nepoch': self.n_epochs_var,
        'configs': self.configs
    }
    if self.is_keras_model:
      input_dicts['configs']['training'] = True
      outputs_train = self._model(**input_dicts)
      input_dicts['configs']['training'] = False
      outputs_test = self._model(**input_dicts)
    # ====== ODIN ====== #
    else:
      all_nnops = N.get_all_nnops(scope=self.name)
      for op in all_nnops:
        if isinstance(op, N.Container):
          op.set_debug(is_verbose())
      # applying to get the output
      outputs_train = self._model(**input_dicts)
      outputs_test = outputs_train
    # ====== compulsory outputs ====== #
    assert sorted(list(outputs_train.keys())) == sorted(list(outputs_test.keys())),\
    "Output for training and testing must be the same set of Tensor"
    assert 'Z' in outputs_train and 'Z' in outputs_test, \
    "Latent space must be in the outputs"
    assert 'loss' in outputs_train and 'loss' in outputs_test, \
    "Loss must be given in the outputs for training"
    if self.is_semi_supervised:
      assert 'y' in outputs_test, \
      "Require output `y` if model is semi-supervised"

    for output_idx, (output_type, outputs) in enumerate(
        zip(('train', 'test'), (outputs_train, outputs_test))):
      # ====== Custom outputs ====== #
      for k, v in outputs.items():
        if k not in possible_outputs:
          self._other_outputs[output_type][k] = v
      # ====== latent space output ====== #
      Z = outputs['Z']
      Z_sample = outputs.get('Z_sample', None)
      assert Z.get_shape().as_list()[-1] == int(self.zdim), \
          "Expect %d latent dimension but return Z with shape: %s" % \
          (int(self.zdim), Z.get_shape().as_list())
      # ====== Loss and metrics ====== #
      # make sure the loss is named loss
      loss = copy_keras_metadata(outputs['loss'],
                                 tf.identity(outputs['loss'], name='loss'))
      if loss.shape.ndims != 0:
        raise RuntimeError("loss must be scalar, but given: %s" % str(loss))
      metr = outputs.get('metr', [])
      if K.is_tensor(metr, inc_distribution=False):
        metr = [metr]
      assert isinstance(metr, list)
      # all metrics must be scalar
      for m in metr:
        if m.shape.ndims != 0:
          raise RuntimeError(
              "Returned metrics must be scalar, but given: %s" % str(m))
      # ====== library size ====== #
      L_expected = outputs.get('L', None)
      L_sample = outputs.get('L_sample', None)
      L_stddev = outputs.get('L_stddev', None)
      # ====== reconstructed ====== #
      W_expected = outputs.get('W', None)
      W_sample = outputs.get('W_sample', None)
      W_stddev = outputs.get('W_stddev', None)
      # ====== imputed ====== #
      V_expected = outputs.get('V', None)
      V_sample = outputs.get('V_sample', None)
      V_stddev = outputs.get('V_stddev', None)
      # ====== zero-inflated pi ====== #
      pi = outputs.get('pi', None)
      # ====== predicted labels ====== #
      y = outputs.get('y', None)
      y_sample = outputs.get('y_sample', None)
      if is_verbose():
        print(ctext("'%s' outputs:" % output_type, 'lightyellow'))
        print("   Latent           :", ctext(Z, 'cyan'))
        print("   Latent (sample)  :", ctext(Z_sample, 'cyan'))
        print("   zero-inflated PI :", ctext(pi, 'cyan'))
        print(ctext("  Labels:", 'lightyellow'))
        print("   Prediction       :", ctext(y, 'cyan'))
        print("   Sample           :", ctext(y_sample, 'cyan'))
        print(ctext("  Library size:", 'lightyellow'))
        print("   L expected       :", ctext(L_expected, 'cyan'))
        print("   L sample         :", ctext(L_sample, 'cyan'))
        print("   L stddev         :", ctext(L_stddev, 'cyan'))
        print(ctext("  Reconstruction:", 'lightyellow'))
        print("   W expected       :", ctext(W_expected, 'cyan'))
        print("   W sample         :", ctext(W_sample, 'cyan'))
        print("   W stddev         :", ctext(W_stddev, 'cyan'))
        print(ctext("  Imputed:", 'lightyellow'))
        print("   V expected       :", ctext(V_expected, 'cyan'))
        print("   V sample         :", ctext(V_sample, 'cyan'))
        print("   V stddev         :", ctext(V_stddev, 'cyan'))
        print(ctext("  Training info:", 'lightyellow'))
        print("   Loss    :", ctext(loss, 'cyan'))
        print("   Metrics :", ctext(metr, 'cyan'))
      self._outputs[output_type] = {
          'z': Z, 'z_sample': Z_sample,
          'pi': pi,
          'y': y, 'y_sample': y_sample,
          'l': L_expected, 'l_sample': L_sample, 'l_stddev': L_stddev,
          'w': W_expected, 'w_sample': W_sample, 'w_stddev': W_stddev,
          'v': V_expected, 'v_sample': V_sample, 'v_stddev': V_stddev,
          'loss': loss,
          'metr': metr
      }
    # initialize everything
    K.initialize_all_variables()

  def _initialize_predict_functions(self):
    if len(self._pred_functions) > 0:
      return self._pred_functions
    # ====== inputs ====== #
    inputs = list(self.input_plhs)[:-2]
    metrics = sorted(
        set(as_list(self._outputs['test']['loss']) +
            as_list(self._outputs['test']['metr'])),
        key=lambda x: x.name)
    self._metrics_name = [_normalize_tensor_name(i.name)
                          for i in metrics]

    def output_iter():
      for name, tensor in chain(self._outputs['test'].items(),
                                self._other_outputs['test'].items()):
        if tensor is None or \
        (isinstance(tensor, (tuple, list)) and len(tensor) == 0):
          continue
        if name in ('loss', 'metr', 'y'):
          continue
        yield name, tensor

    # ====== for Keras ====== #
    if self.is_keras_model:
      from tensorflow.python.keras import Model
      # unsupervised
      for name, tensor in output_iter():
        m = Model(inputs=inputs, outputs=tensor,
                  name="predict_%s" % name)
        if '_sample' in name: # special case for sample
          # NOTE: no batch support (careful memory)
          f = m.predict_on_batch
        else: # other output
          f = partial(m.predict, batch_size=PREDICTION_BATCH_SIZE, verbose=0)
        self._pred_functions[name] = f
      # metrics
      self._pred_functions['metrics'] = partial(
          Model(inputs=[i for i in self.input_plhs if i is not None],
                outputs=metrics,
                name="predict_metrics").predict,
          batch_size=PREDICTION_BATCH_SIZE, verbose=0)
      # semi-supervised
      if self.is_semi_supervised:
        # it is notable that prediction of y must not require
        # y as input data, we want to predict y not denoising y
        m = Model(inputs=inputs,
                  outputs=self._outputs['test']['y'],
                  name="predict_y")
        self._pred_functions['y'] = partial(
            m.predict, batch_size=PREDICTION_BATCH_SIZE, verbose=0)
      return self._pred_functions

    # ====== for ODIN ====== #
    def create_func(ins_, out_, is_scoring=False):
      if isinstance(out_, (tuple, list)):
        out_ = [o for o in out_ if o is not None]
        if len(out_) == 0:
          out_ = None
      if out_ is not None:
        fn = K.function(
            inputs=ins_, outputs=out_, training=False)

        # this function will iterate over mini-batch for prediction
        # make sure the order is preserved, so no shuffling
        def _batch_prediction(X):
          assert len(set(x.shape[0] for x in X)) == 1
          n = X[0].shape[0]
          all_y = []
          for start, end in batching(batch_size=PREDICTION_BATCH_SIZE,
                                     n=n, seed=None):
            x = [i[start:end] for i in X]
            y = fn(*x)
            all_y.append(y)
          # concatenate the results
          if not is_scoring:
            if isinstance(y, (tuple, list)):
              return [np.concatenate([j[i] for j in all_y], axis=0)
                      for i in range(len(y))]
            else:
              return np.concatenate(all_y, axis=0)
          # scoring mode (return scalar)
          else:
            return np.mean(np.array(all_y), axis=0)
        return _batch_prediction
      return None

    # unsupervised
    for name, tensor in output_iter():
      self._pred_functions[name] = create_func(ins_=inputs, out_=tensor)
    # metrics
    self._pred_functions['metrics'] = create_func(
        ins_=[i for i in self.input_plhs if i is not None],
        out_=metrics,
        is_scoring=True)
    # semi-supervised
    if self.is_semi_supervised:
      # it is notable that prediction of y must not require
      # y as input data, we want to predict y not denoising y
      self._pred_functions['y'] = create_func(
          ins_=inputs,
          out_=self._outputs['test']['y'])
    return self._pred_functions

  # ******************** helper ******************** #
  def normalize(self, x, method, data_name):
    x = np.atleast_2d(x)
    if method == 'raw':
      pass
    elif method == 'log':
      x = K.log_norm(x, axis=1, scale_factor=10000)
    elif method == 'bin' or method == 'prob':
      is_binary_classes = sorted(np.unique(x.astype('float32'))) == [0., 1.]
      if not is_binary_classes:
        if data_name not in self._gmm_threshold:
          gmm = ProbabilisticEmbedding()
          gmm.fit(x)
          self._gmm_threshold[data_name] = gmm
        else:
          gmm = self._gmm_threshold[data_name]
        x = gmm.predict(x) if method == 'bin' else gmm.predict_proba(x)
      else: # already binarized or probabilized
        pass
    else:
      raise NotImplementedError
    return x

  def _preprocess_inputs(self, X, y):
    """
    Return
    ------
    X_norm, T_norm, L, L_mean, L_var, y_norm

    Note
    ----
    y_norm could be None
    """
    assert X.shape[1] == self.gene_dim, \
    "This model require %d gene expression, but given X.shape=%s" % \
    (self.gene_dim, str(X.shape))

    if self.is_semi_supervised and y is not None:
      assert y.shape[1] == self.prot_dim,\
      "This model require %d protein markers, but given y.shape=%s" % \
      (self.prot_dim, str(y.shape))
    # ====== gene expression ====== #
    if X.ndim == 1:
      X = np.expand_dims(X, axis=-1)
    if self.xclip > 0:
      X = np.clip(X, a_min=0, a_max=float(self.xclip))
    # ====== library size ====== #
    L, L_mean, L_var = get_library_size(X, return_library_size=True)
    X_norm = self.normalize(X, method=self.xnorm, data_name='X')
    T_norm = self.normalize(X, method=self.tnorm, data_name='X')
    assert self._gene_dim == X_norm.shape[1], "Number of genes mismatch"
    # ====== protein ====== #
    y_norm = None
    if self.is_semi_supervised and y is not None:
      if y.ndim == 1:
        y = np.expand_dims(y, axis=-1)
      if self.yclip > 0:
        y = np.clip(y, a_min=0, a_max=float(self.yclip))

      y_norm = self.normalize(y, method=self.ynorm, data_name='y')
      assert X_norm.shape[0] == y_norm.shape[0]
    # ====== return ====== #
    return X_norm, T_norm, L, L_mean, L_var, y_norm

  # ******************** fitting ******************** #
  def fit(self, X, y=None,
          supervised_percent=0.8, validation_percent=0.1,
          n_mcmc_samples=1,
          corruption_rate=0.25, corruption_dist='binomial',
          batch_size=64, n_epoch=120, learning_rate=1e-4,
          monitoring=False, fig_dir=None,
          detail_logging=False):
    """

    Parameters
    ----------
    X : [n_samples, n_genes]
        single-cell gene expression matrix

    y : [n_samples, n_protein]
        single-cell protein marker matrix, if not given,
        assume to be the same as `X`, and might not be used
        by the model

    supervised_percent : float (0. to 1.)
        percent of training data used for supervised objective

    validation_percent : float (0. to 1.)
        percent of training data used for validation at the
        end of each epoch, if 0. then no validation is performed

    n_mcmc_samples : int (> 0)
        number of MCMC samples for training

    batch_size : int (default: 64)
        batch size for training

    n_epoch : int (default: 120)
        number of epoch for training

    learning_rate : float (default: 0.0001)
        learning rate for Adam

    monitoring : bool
        enable monitoring each epoch

    fig_dir : {None, string}
        path for saving the training summary figure,
        if not given, then no training summary is stored

    detail_logging : bool (default: False)
        if True, print loss of the monitoring metrics
        during training, otherwise, only so the current
        epoch progress

    """
    training_log = io.StringIO()
    stdio(training_log)

    if self._corruption_rate is None:
      self._corruption_rate = corruption_rate
      self._corruption_dist = corruption_dist
    else:
      assert (self._corruption_rate == corruption_rate and
              self._corruption_dist == corruption_dist),\
      ("Model have been fitted with following corruption configurations: %s, %s;"
       "but given: %s, %s" % (self._corruption_rate, self._corruption_dist,
                              corruption_rate, corruption_dist))
    n_samples = X.shape[0]
    self._ps = supervised_percent
    self._batch_size = batch_size
    self._learning_rate = learning_rate
    self._n_mcmc_train = n_mcmc_samples
    # apply artificial corruption for training
    X = apply_artificial_corruption(
        X, dropout=corruption_rate, distribution=corruption_dist)
    # ====== check semi-supervised system ====== #
    if self.is_semi_supervised and y is None:
      raise ValueError(
          "`y` must be not None for fitting semi-supervised model")
    X, T, L, L_mean, L_var, y = self._preprocess_inputs(X, y)
    # ====== splitting train valid ====== #
    (X_train, T_train,
     L_train, Lmean_train, Lvar_train,
     y_train) = None, None, None, None, None, None
    (X_valid, T_valid,
     L_valid, Lmean_valid, Lvar_valid,
     y_valid) = None, None, None, None, None, None

    if validation_percent > 0:
      train_ids, valid_ids = train_valid_test_split(
          x=np.arange(n_samples, dtype='int32'),
          train=1 - validation_percent,
          inc_test=False,
          seed=UNIVERSAL_RANDOM_SEED)
      X_train = X[train_ids]
      X_valid = X[valid_ids]

      T_train = T[train_ids]
      T_valid = T[valid_ids]

      L_train = L[train_ids]
      L_valid = L[valid_ids]
      Lmean_train = L_mean[train_ids]
      Lmean_valid = L_mean[valid_ids]
      Lvar_train = L_var[train_ids]
      Lvar_valid = L_var[valid_ids]

      if y is not None:
        y_train = y[train_ids]
        y_valid = y[valid_ids]
    else: # no validation set given
      X_train = X
      T_train = T
      L_train = L
      Lmean_train = L_mean
      Lvar_train = L_var
      if y is not None:
        y_train = y

    n_train = len(X_train)
    n_valid = 0 if X_valid is None else len(X_valid)
    # ====== supervised mask ====== #
    # Generate a mask of 0, 1
    # : 0 mean unsupervised sample,
    # : 1 mean supervised sample
    rand = np.random.RandomState(seed=UNIVERSAL_RANDOM_SEED)
    m_train = np.zeros(shape=n_train, dtype='float32')
    supervised_indices = rand.choice(
        np.arange(0, n_train),
        size=int(supervised_percent * n_train),
        replace=False)
    m_train[supervised_indices] = 1
    assert int(np.sum(m_train)) == int(supervised_percent * n_train)
    # we also need mask for validation but take into acocunt all
    # validation sample to evaluate supervised tasks
    m_valid = np.ones(shape=(n_valid,), dtype='float32')
    # ====== prepare the placeholders ====== #
    inputs = [i for i in self.input_plhs
              if i is not None]
    train_data = (X_train, T_train,
                  L_train, Lmean_train, Lvar_train,
                  m_train, y_train)
    valid_data = None if n_valid is 0 else\
    (X_valid, T_valid,
     L_valid, Lmean_valid, Lvar_valid,
     m_valid, y_valid)
    if not self.is_semi_supervised:
      train_data = train_data[:-2]
      if valid_data is not None:
        valid_data = valid_data[:-2]
    # ====== set value of the variable ====== #
    assert int(n_mcmc_samples) >= 1, \
    "Number of MCMC samples must be >= 1"
    K.set_value(self.n_mcmc_samples_var, int(n_mcmc_samples))
    # ====== training for keras ====== #
    if self.is_keras_model:
      from tensorflow.python.keras import Model
      from tensorflow.python.keras.optimizers import Adam
      if self.is_semi_supervised:
        model = Model(inputs=inputs,
          outputs=KerasConcatenate(axis=-1)([self._outputs['train']['w'],
                                             self._outputs['train']['y']]))
      else:
        model = Model(inputs=inputs,
                      outputs=self._outputs['train']['w'])
      optz = Adam(lr=learning_rate)

      if is_verbose():
        print(ctext("Trainable variables:", 'lightyellow'))
        for v in sorted(model.trainable_variables, key=lambda x: x.name):
          print(" ", v.name, v.shape)
        print(ctext("None-trainable variables:", 'lightyellow'))
        for v in sorted(model.non_trainable_variables, key=lambda x: x.name):
          print(" ", v.name, v.shape)

      model.compile(optz,
          loss=to_keras_objective(self._outputs['train']['loss']),
          metrics=[to_keras_objective(i)
                 for i in self._outputs['train']['metr']])
      hist = model.fit(x=train_data, y=None,
          batch_size=batch_size, epochs=n_epoch,
          verbose=1 if detail_logging else 0,
          callbacks=[LambdaCallback(
              on_epoch_begin=lambda *args, **kwargs:
              K.set_value(self.n_epochs_var, self.n_epochs_var + 1))],
          validation_data=(valid_data, None),
          shuffle=True)
      # extract the history
      for k, v in hist.history.items():
        name = k
        hist_type = 'train'
        if 'val_' == k[:4]:
          name = k[4:]
          hist_type = 'valid'
        # normalize all the tensor name
        self._history[hist_type][_normalize_tensor_name(name)] += list(v)
    # ====== training for O.D.I.N ====== #
    else:
      # create the optimizer
      optz = K.optimizers.Adam(lr=float(learning_rate), clipnorm=None,
                               name=self.name)
      updates = optz.minimize(self._outputs['train']['loss'],
                              verbose=is_verbose())
      global_norm = optz.norm
      K.initialize_all_variables()
      # ====== training functions ====== #
      # for training
      tmp = [self._outputs['train']['loss'], global_norm]
      for i in self._outputs['train']['metr']:
        if i not in tmp:
          tmp.append(i)
      f_train = K.function(inputs=inputs, outputs=tmp, training=True,
                           updates=updates)
      # for scoring
      tmp = [self._outputs['train']['loss']]
      for i in self._outputs['train']['metr']:
        if i not in tmp:
          tmp.append(i)
      f_score = K.function(inputs=inputs, outputs=tmp, training=False,
                           batch_size=None)
      # overlapping metrics for summary
      overlap_metrics = sorted(set([self._outputs['train']['loss']] +
                                   self._outputs['train']['metr']),
                               key=lambda x: x.name)
      if is_verbose():
        print(ctext("Metrics for monitoring:", 'lightyellow'))
        for m in overlap_metrics:
          print("  %s:" % ctext(m.name, 'cyan'), m)
      # ====== create training loop ====== #
      trainer = training.MainLoop(batch_size=int(batch_size),
                                  seed=UNIVERSAL_RANDOM_SEED, shuffle_level=2,
                                  allow_rollback=False,
                                  verbose=2 if bool(detail_logging) else 3,
                                  labels=None)
      trainer.set_checkpoint(obj=self._model)
      # ====== callbacks ====== #
      all_callbacks = [
          training.LambdaCallback(
              fn=lambda:K.set_value(self.n_epochs_var, self.n_epochs_var + 1),
              task_name='train',
              signal=training.TaskSignal.EpochStart),
          training.NaNDetector(),
          training.CheckpointGeneralization('valid',
                                      output_name=self._outputs['train']['loss']),
          # epoch summaries
          None if fig_dir is None else LearningCurves(save_path=fig_dir),
          # monitoring latent spaces
          training.LambdaCallback(fn=self.monitoring_epoch, task_name='train',
                                  signal=training.TaskSignal.EpochEnd)
          if bool(monitoring) else None,
      ]
      trainer.set_callbacks(all_callbacks)
      # ====== training task ====== #
      trainer.set_train_task(func=f_train,
                             data=train_data,
                             epoch=int(n_epoch),
                             name='train')
      if X_valid is not None:
        trainer.set_valid_task(func=f_score,
                               data=valid_data,
                               freq=training.Timer(percentage=1.0),
                               name='valid')
      # NOTE: this line is important
      trainer.run()
      # ====== store the history ====== #
      hist = trainer.history
      for task_name, task_results in hist.items():
        for epoch_id, epoch_results in task_results.items():
          for tensor_name, tensor_batch in epoch_results.items():
            tensor_name = _normalize_tensor_name(tensor_name)
            self._history[task_name][tensor_name].append(np.mean(tensor_batch))
    # ====== end training update states ====== #
    self._trained_n_epoch += n_epoch
    self._is_fitted = True
    stdio()
    training_log.seek(0)
    self._training_log += '\n%s\n' % get_formatted_datetime(only_number=False)
    self._training_log += training_log.read()
    training_log.close()

  # ******************** history ******************** #
  def get_train_loss(self):
    return self._history['train']['loss']

  def get_valid_loss(self):
    return self._history['valid']['loss']

  def get_train_history(self, name):
    return self._history['train'][name]

  def get_valid_history(self, name):
    return self._history['valid'][name]

  # ******************** plotting utils ******************** #
  def plot_tsne(self, X, y=None,
                labels=None, labels_name=None,
                n_samples=1000, n_mcmc_samples=100, output_type='Z',
                show_pca=False, ax=None, title=None, seed=5218):
    """
    Parameters
    ----------
    output_type : string
        Z : latent space
        W : reconstructed gene expression
        V : denoised gene expression
        PI : Zero-inflated rate
    """
    X, T, C, y = self._preprocess_inputs(X, y)
    if labels is None:
      labels = np.ones(shape=(X.shape[0], 1))
    if labels.ndim == 1:
      labels = np.expand_dims(labels, axis=-1)
    if np.max(labels) > 1 and labels.shape[1] == 1:
      labels = one_hot(y=labels.ravel(),
        nb_classes=np.max(labels) + 1
        if labels_name is None else
        len(labels_name))
    assert len(labels) == X.shape[0], \
    "Number of samples mismatch between X and labels"

    n_classes = labels.shape[1]
    if labels_name is None:
      labels_name = np.array(['#%d' % i for i in range(n_classes)])
    # ====== downsampling ====== #
    rand = np.random.RandomState(seed=seed)
    if n_samples is not None:
      ids = rand.permutation(X.shape[0])
      ids = rand.choice(ids, size=n_samples, replace=False)
      X = X[ids]
      T = T[ids]
      C = C[ids]
      y = y[ids]
      if labels is not None:
        labels = labels[ids]
    # ====== make prediction ====== #
    f = self._initialize_predict_functions()
    mask = np.ones(shape=(X.shape[0],))
    Z = f[str(output_type).lower()](X, T, C, mask, y)
    if isinstance(Z, (tuple, list)):
      Z = Z[0]
    if Z is None:
      raise RuntimeError(
          "Model has no support for output type: '%s'" % output_type)
    # ====== apply pca and tsne ====== #
    is_single_label = np.allclose(np.max(labels, axis=1), 1)
    Z_pca = fast_pca(Z, n_components=min(512, Z.shape[1]), algo='pca',
                     random_state=rand.randint(10e8))
    if not bool(show_pca):
      Z_tsne = fast_tsne(Z_pca, n_components=2,
                         random_state=rand.randint(10e8))
    else:
      Z_tsne = None
    # ====== plot ====== #
    from odin.visual import (plot_scatter, plot_figure, plot_scatter_heatmap)
    if is_single_label:
      plot_scatter(x=Z_pca[:, :2] if bool(show_pca) else Z_tsne,
                   color=[labels_name[i] for i in np.argmax(labels, axis=-1)],
                   size=6, ax=ax, grid=False,
                   legend_enable=True, legend_ncol=4, fontsize=8,
                   title=title)
    else:
      colormap = 'Reds' # bwr
      ncol = 5 if n_classes <= 20 else 9
      nrow = int(np.ceil(n_classes / ncol))
      fig = plot_figure(nrow=4 * nrow, ncol=20)
      for i, name in enumerate(labels_name):
        val = K.log_norm(labels[:, i], axis=0)
        plot_scatter_heatmap(
            x=Z_pca[:, :2] if bool(show_pca) else Z_tsne,
            val=val / np.sum(val),
            ax=(nrow, ncol, i + 1),
            colormap=colormap, size=8, alpha=0.8,
            fontsize=8, grid=False,
            title='[%s]%s' % ("PCA" if show_pca else "t-SNE", name))

      import matplotlib as mpl
      cbar_ax = fig.add_axes([0.92, 0.15, 0.02, 0.7])
      cmap = mpl.cm.get_cmap(name=colormap)
      norm = mpl.colors.Normalize(vmin=0., vmax=1.)
      cb1 = mpl.colorbar.ColorbarBase(cbar_ax, cmap=cmap,
                                      norm=norm,
                                      orientation='vertical')
      cb1.set_label('Protein markers level')

  # ******************** predicting ******************** #
  def _make_prediction(self, pred_type, X, y=None, n_mcmc_samples=100):
    K.set_value(self.n_mcmc_samples_var, n_mcmc_samples)
    X_norm, T_norm, L, L_mean, L_var, y_norm = self._preprocess_inputs(X, y)
    mask = np.ones(shape=(X.shape[0],))
    f = self._initialize_predict_functions()
    key = str(pred_type).lower()
    if key not in f:
      raise RuntimeError("No prediction for '%s'" % pred_type)
    f = f[key]
    if y is None:
      return f((X_norm, T_norm, L, L_mean, L_var))
    else:
      return f((X_norm, T_norm, L, L_mean, L_var, mask, y))

  # ******************** scoring ******************** #
  def score(self, X, y=None, n_mcmc_samples=100, return_mean=True):
    """Compute the per-sample average of recorded metrics of the given data X.

    Parameters
    ----------
    X : array-like, shape (n_samples, n_dimensions)
        List of n_features-dimensional data points. Each row
        corresponds to a single data point.

    Returns
    -------
    dictionary : metric_name -> values
    """
    if not self.is_semi_supervised:
      y = None
    scores = self._make_prediction(
        'metrics', X=X, y=y, n_mcmc_samples=n_mcmc_samples)
    if self.is_keras_model:
      scores = [np.mean(i) for i in scores]
    return {_normalize_tensor_name(name): s
            for name, s in zip(self._metrics_name, scores)}

  # ******************** sampling function ******************** #
  def _sample(self, name, X, n_mcmc_samples):
    output = []
    for s, e in batching(batch_size=PREDICTION_BATCH_SIZE, n=X.shape[0]):
      x = X[s:e]
      output.append(
          self._make_prediction(pred_type=name, X=x, n_mcmc_samples=n_mcmc_samples))
    return np.concatenate(output, axis=1)

  def sample_Z(self, X, n_mcmc_samples=1):
    return self._sample('Z_sample', X, n_mcmc_samples)

  def sample_y(self, X, n_mcmc_samples=1):
    assert self.is_semi_supervised,\
    "Only semi-supervised model has this prediction!"
    return self._sample('y_sample', X, n_mcmc_samples)

  def sample_W(self, X, n_mcmc_samples=1):
    return self._sample('W_sample', X, n_mcmc_samples)

  def sample_V(self, X, n_mcmc_samples=1):
    return self._sample('V_sample', X, n_mcmc_samples)

  def sample_L(self, X, n_mcmc_samples=1):
    try:
      return self._sample('L_sample', X, n_mcmc_samples)
    except RuntimeError:
      return np.sum(self.sample_V(X, n_mcmc_samples), axis=-1, keepdims=True)

  # ******************** predicting ******************** #
  def predict_Z(self, X):
    """ Return mean of the latent posterior
    (i.e. mean of Normal distribution) """
    return self._make_prediction('Z', X, n_mcmc_samples=1)

  def predict_y(self, X, n_mcmc_samples=100):
    """ Predicting the protein marker """
    # and y is not None
    assert self.is_semi_supervised,\
    "Only semi-supervised model has this prediction!"
    return self._make_prediction(
        pred_type='y', X=X, y=None, n_mcmc_samples=n_mcmc_samples)

  def predict_L(self, X, n_mcmc_samples=100):
    """ Return library size prediction """
    try:
      return self._make_prediction('L', X, n_mcmc_samples=n_mcmc_samples)
    except RuntimeError:
      return np.sum(self._make_prediction('V', X, n_mcmc_samples=n_mcmc_samples),
                    axis=-1, keepdims=True)

  def predict_Lstddev(self, X, n_mcmc_samples=100):
    """ Return library size stddev prediction """
    try:
      return self._make_prediction('L_stddev', X, n_mcmc_samples=n_mcmc_samples)
    except RuntimeError:
      return np.sum(self._make_prediction('V_stddev', X, n_mcmc_samples=n_mcmc_samples),
                    axis=-1, keepdims=True)

  def predict_W(self, X, n_mcmc_samples=100):
    """ Return mean of the reconstructed value """
    return self._make_prediction('W', X, n_mcmc_samples=n_mcmc_samples)

  def predict_Wstddev(self, X, n_mcmc_samples=100):
    """ Return stddev of reconstructed value
    if not a variational model, then V_stddev is None
    """
    return self._make_prediction('W_stddev', X, n_mcmc_samples=n_mcmc_samples)

  def predict_V(self, X, n_mcmc_samples=100):
    """ Return mean of the denoised value """
    return self._make_prediction('V', X, n_mcmc_samples=n_mcmc_samples)

  def predict_Vstddev(self, X, n_mcmc_samples=100):
    """ Return stddev of denoised value
    if not a variational model, then V_stddev is None
    """
    return self._make_prediction('V_stddev', X, n_mcmc_samples=n_mcmc_samples)

  def predict_PI(self, X, n_mcmc_samples=100):
    """ Return a matrix (n_sample, n_gene) of Zero-inflated
    rate

    if not a zero-inflated model, then return None
    """
    return self._make_prediction('PI', X, n_mcmc_samples=n_mcmc_samples)

  def predict(self, X, n_mcmc_samples=100):
    """ Return denoised values, same as `predict_V` """
    return self.predict_V(X, n_mcmc_samples)

  # ******************** properties ******************** #
  @property
  def id(self):
    """ Unique ID for this inference class """
    assert self.is_fitted, "Only fitted Inference has ID"
    name = self.configs['module_name'] if 'module_name' in self.configs else \
    self.name
    name = name.strip().lower()

    kl_weight_text = '%.2f' % self.kl_weight
    kl_weight_text = kl_weight_text.split('.')
    if int(kl_weight_text[1]) == 0:
      kl_weight_text[1] = ''
    kl_weight_text = kl_weight_text[0] + kl_weight_text[1][:2]

    y_weight = '%.2g' % self.configs['y_weight']
    y_weight = y_weight.replace('.', '')

    model_id = '_'.join([
        name,
        'gene' if self.dispersion == 'gene' else 'genecell',
        'X%s%d%s' % (self.xnorm, int(max(self.xclip, 0)), self.xdist),
        'Y%s%d%s' % (self.ynorm, int(max(self.yclip, 0)), self.ydist),
        'T%s' % self.tnorm,
        'Z' + str(self.zdist),
        '%.3dspvs' % (self.supervised_percent * 100) + y_weight,
        'net%.2d%.3d%.3d' % (self.nlayer, self.hdim, self.zdim),
        'drop%.2d%.2d%.2d%.2d' % (self.xdrop * 100,
                                  self.edrop * 100,
                                  self.zdrop * 100,
                                  self.ddrop * 100),
        '%skl%d' % (kl_weight_text, int(self.warmup)),
        'alytcT' if bool(self.analytic) else 'alytcF',
        'bnormT' if bool(self.batchnorm) else 'bnormF',
        '%s%.2d' % (self.corruption_dist, self.corruption_rate * 100)
    ])
    return model_id

  @property
  def short_id(self):
    """ Shorter id, remove all the training configuration """
    return '_'.join(self.id.split('_')[:-5])

  @property
  def kl_weight(self):
    return float(self._configs['kl_weight'])

  @property
  def warmup(self):
    """ Number of warmup epoch for KL divergence """
    return float(self._configs['warmup'])

  @property
  def name(self):
    return self._name

  @property
  def nlayer(self):
    return self._configs['nlayer']

  @property
  def hdim(self):
    """ Number of hidden units for intermediate layers """
    return self._configs['hdim']

  @property
  def zdim(self):
    """ Number of hidden units for the bottleneck layer """
    return self._configs['zdim']

  @property
  def inputs(self):
    return [i for i in self.input_plhs if i is not None]

  @property
  def outputs(self):
    o = {}
    o['train'] = dict(self._outputs['train'])
    o['test'] = dict(self._outputs['test'])
    o['train'].update(self._other_outputs['train'])
    o['test'].update(self._other_outputs['test'])
    return o

  @property
  def corruption_rate(self):
    return self._corruption_rate

  @property
  def corruption_dist(self):
    return self._corruption_dist

  @property
  def configs(self):
    return dict(self._configs)

  @property
  def history(self):
    return copy.deepcopy(self._history)

  @property
  def is_keras_model(self):
    return self._configs['is_keras_model']

  @property
  def dispersion(self):
    return self._configs['dispersion']

  @property
  def is_semi_supervised(self):
    return self._prot_dim > 0

  @property
  def supervised_percent(self):
    return self._ps

  @property
  def xclip(self):
    return self._configs['xclip']

  @property
  def xdrop(self):
    return self._configs['xdrop']

  @property
  def edrop(self):
    return self._configs['edrop']

  @property
  def zdrop(self):
    return self._configs['zdrop']

  @property
  def ddrop(self):
    return self._configs['ddrop']

  @property
  def yclip(self):
    return self._configs['yclip']

  @property
  def xnorm(self):
    return self._configs['xnorm']

  @property
  def xdist(self):
    return self._configs['xdist']

  @property
  def ydist(self):
    return self._configs['ydist']

  @property
  def zdist(self):
    return self._configs['zdist']

  @property
  def tnorm(self):
    return self._configs['tnorm']

  @property
  def ynorm(self):
    return self._configs['ynorm']

  @property
  def y_weight(self):
    return self._configs['y_weight']

  @property
  def is_fitted(self):
    return self._is_fitted

  @property
  def gene_dim(self):
    """ Return number of genes used for training """
    return self._gene_dim

  @property
  def prot_dim(self):
    """ Return number of protein markers used for training """
    return self._prot_dim

  @property
  def analytic(self):
    return self._configs['analytic']

  @property
  def batchnorm(self):
    return self._configs['batchnorm']

  @property
  def training_log(self):
    return self._training_log

# ===========================================================================
# Shortcut
# ===========================================================================
class InferenceSCVAE(Inference):
  def __init__(self, gene_dim, dispersion='gene-cell',
               xnorm='log', tnorm='raw', ynorm='prob',
               xclip=0, yclip=0,
               xdist='zinb', ydist='bernoulli', zdist='normal',
               xdrop=0.3, edrop=0, zdrop=0, ddrop=0,
               hdim=128, zdim=32, nlayer=2,
               batchnorm=True, analytic=True,
               kl_weight=1., warmup=400, y_weight=10.,
               extra_module_path=None,
               **kwargs):
    super(InferenceSCVAE, self).__init__(gene_dim=gene_dim, prot_dim=None,
               model='vae', dispersion=dispersion,
               xnorm=xnorm, tnorm=tnorm, ynorm=ynorm,
               xclip=xclip, yclip=yclip,
               xdist=xdist, ydist=ydist, zdist=zdist,
               xdrop=xdrop, edrop=edrop, zdrop=zdrop, ddrop=ddrop,
               hdim=hdim, zdim=zdim, nlayer=nlayer,
               batchnorm=batchnorm, analytic=analytic,
               kl_weight=kl_weight, warmup=warmup, y_weight=y_weight,
               extra_module_path=extra_module_path,
               **kwargs)

class InferenceSISUA(Inference):
  def __init__(self, gene_dim, prot_dim,
               dispersion='gene-cell',
               xnorm='log', tnorm='raw', ynorm='prob',
               xclip=0, yclip=0,
               xdist='zinb', ydist='bernoulli', zdist='normal',
               xdrop=0.3, edrop=0, zdrop=0, ddrop=0,
               hdim=128, zdim=32, nlayer=2,
               batchnorm=True, analytic=True,
               kl_weight=1., warmup=400, y_weight=10.,
               extra_module_path=None,
               **kwargs):
    super(InferenceSISUA, self).__init__(gene_dim=gene_dim, prot_dim=prot_dim,
               model='movae', dispersion=dispersion,
               xnorm=xnorm, tnorm=tnorm, ynorm=ynorm,
               xclip=xclip, yclip=yclip,
               xdist=xdist, ydist=ydist, zdist=zdist,
               xdrop=xdrop, edrop=edrop, zdrop=zdrop, ddrop=ddrop,
               hdim=hdim, zdim=zdim, nlayer=nlayer,
               batchnorm=batchnorm, analytic=analytic,
               kl_weight=kl_weight, warmup=warmup, y_weight=y_weight,
               extra_module_path=extra_module_path,
               **kwargs)



from __future__ import absolute_import, division, print_function

import base64
import os
import pickle
import shutil
from io import BytesIO, StringIO

import numpy as np
import scipy as sp

from odin.fuel import Dataset
from odin.utils import batching, crypto, ctext, get_file
from sisua.data.path import DOWNLOAD_DIR, PREPROCESSED_BASE_DIR
from sisua.data.utils import remove_allzeros_columns, save_to_dataset

_CITEseq_CBMC_PREPROCESSED = os.path.join(PREPROCESSED_BASE_DIR,
                                          'CBMC_preprocessed')

_URL = b'aHR0cHM6Ly9zMy5hbWF6b25hd3MuY29tL2FpLWRhdGFzZXRzL0NCTUMuemlw\n'
_PASSWORD = 'uef-czi'


def read_CITEseq_CBMC(override=False, verbose=False):
  download_path = os.path.join(DOWNLOAD_DIR, "CBMC_original")
  if not os.path.exists(download_path):
    os.mkdir(download_path)

  preprocessed_path = _CITEseq_CBMC_PREPROCESSED
  if not os.path.exists(preprocessed_path):
    os.mkdir(preprocessed_path)
  elif override:
    if verbose:
      print("Overriding path: %s" % _CITEseq_CBMC_PREPROCESSED)
    shutil.rmtree(_CITEseq_CBMC_PREPROCESSED)
    os.mkdir(_CITEseq_CBMC_PREPROCESSED)
  # ******************** preprocessed data NOT found ******************** #
  if not os.path.exists(os.path.join(preprocessed_path, 'X')):
    X, X_row, X_col = [], None, None
    y, y_row, y_col = [], None, None
    # ====== download the data ====== #
    url = str(base64.decodebytes(_URL), 'utf-8')
    base_name = os.path.basename(url)
    get_file(fname=base_name, origin=url, outdir=download_path, verbose=verbose)
    zip_path = os.path.join(download_path, base_name)
    # ====== extract the data ====== #
    data_dict = {}
    for name, data in crypto.unzip_aes(zip_path,
                                       password=_PASSWORD,
                                       verbose=False):
      base_name = os.path.splitext(name)[0]
      if '.npz' in name:
        data = sp.sparse.load_npz(BytesIO(data)).todense()
      elif '.csv' in name:
        data = np.loadtxt(StringIO(str(data, 'utf-8')),
                          dtype=str,
                          delimiter=',')
      else:
        raise RuntimeError("Unknown format: %s" % name)
      data_dict[base_name] = data
    # ====== post-processing ====== #
    X = np.array(data_dict['X'].astype('float32'))
    X_row, X_col = data_dict['X_row'], data_dict['X_col']
    X, X_col = remove_allzeros_columns(matrix=X, colname=X_col)
    assert len(X_row) == X.shape[0] and len(X_col) == X.shape[1]

    y = data_dict['y'].astype('float32')
    y_row, y_col = data_dict['y_row'], data_dict['y_col']
    assert len(y_row) == y.shape[0] and len(y_col) == y.shape[1]

    assert np.all(X_row == y_row), \
    "Cell order mismatch between gene count and protein count"

    # save data
    if verbose:
      print("Saving data to %s ..." % ctext(preprocessed_path, 'cyan'))

    save_to_dataset(preprocessed_path,
                    X,
                    X_col,
                    y,
                    y_col,
                    rowname=X_row,
                    print_log=verbose)
  # ====== read preprocessed data ====== #
  ds = Dataset(preprocessed_path, read_only=True)
  return ds



from collections import OrderedDict

UNIVERSAL_RANDOM_SEED = 5218

# Choose 3D or 2D visualization here
TSNE_DIM = 2

# This contains pair of protein markers that
# should not appear together, hence, good for
# inspecting the biological plausibility of the model
PROTEIN_PAIR_COMPARISON = [
    ('CD8', 'CD4'),
    ('CD56', 'NKT'),
    ('CD45RA', 'CD45RO'),
    ('TIGIT', 'CD127'),
]

# Mapping from protein to gene expression
# CCR5
# CCR7
MARKER_GENES = {
    ### For PBMC CITEseq
    'CD14': 'CD14',
    'CD15': 'FUT4',
    'CD16': 'FCGR3A',
    'CD11c': 'ITGAX',
    'CD127': 'IL7R',
    'CD19': 'CD19',
    'CD2': 'CD2',
    'CD25': 'IL2RA',
    'CD3': 'CD3G',
    'CD4': 'CD4',
    'CD45RA': 'PTPRC',
    'CD45RO': 'PTPRC',
    'CD56': 'NCAM1',
    'CD57': 'B3GAT1',
    'CD8': 'CD8A',
    'CD8a': 'CD8A',
    'PD-1': 'PDCD1',
    'TIGIT': 'TIGIT',
    ### For FACS_5
    'CD20': 'MS4A1',
    'CD45': 'PTPRC',
    'CD34': 'CD34',
    'CD10': 'MME',
    # CD19 is above
    ### For FACS_7
    # CD10 is above
    'CD135': 'FLT3',
    # CD34 is above
    'CD38': 'CD38',
    # CD45RA is above
    'CD49F': 'ITGA6',
    'CD90': 'THY1',
}

# ['CD77' 'CCR7' 'CD19' 'CD1a' 'CD2' 'CD27' 'PD-L1;CD274' 'CD28'
#  'PECAM;CD31' 'CD34' 'CD3' 'CD4' 'CD44' 'CD5' 'CD69' 'CD7' 'CD8' 'CD66b'
#  'CTLA4' 'CD26;Adenosine' 'CD16' 'CD366;tim3' 'HLA-A' 'MHCII;HLA-DR'
#  'IL7Ralpha;CD127' 'CD11b' 'CD11c' 'LAMP1' 'CD56' 'PD-1;CD279' 'PD1;CD279'
#  'B220;CD45R' 'CD45RA' 'CD45RO' 'CD138' 'CD62L' 'Siglec-8' 'Ox40;CD134']



# Modfied code from scVI to SISUA for loading 10xGenomics datasets
#
# MIT License
# Copyright (c) 2018, Romain Lopez
# ...
#
# Full license: https://github.com/YosefLab/scVI/blob/master/LICENSE
# Original code: https://github.com/YosefLab/scVI/blob/master/scvi/dataset/dataset10X.py
#
# For more information about how to read data from 10xGenomic repos:
# https://support.10xgenomics.com/single-cell-gene-expression/software/pipelines/latest/output/matrices
from __future__ import absolute_import, division, print_function

import gzip
import os
import pickle
import shutil
import tarfile
from typing import Tuple

import numpy as np
import pandas as pd
from scipy.io import mmread
from scipy.sparse import csr_matrix, issparse

from odin.fuel import Dataset
from odin.utils import batching, ctext, get_file, is_gzip_file, select_path
from sisua.data.path import DOWNLOAD_DIR, PREPROCESSED_BASE_DIR
from sisua.data.utils import remove_allzeros_columns, save_to_dataset

available_datasets = {
    "1.1.0": [
        "frozen_pbmc_donor_a",
        "frozen_pbmc_donor_b",
        "frozen_pbmc_donor_c",
        "fresh_68k_pbmc_donor_a",
        "cd14_monocytes",
        "b_cells",
        "cd34",
        "cd56_nk",
        "cd4_t_helper",
        "regulatory_t",
        "naive_t",
        "memory_t",
        "cytotoxic_t",
        "naive_cytotoxic",
    ],
    "2.1.0": ["pbmc8k", "pbmc4k", "t_3k", "t_4k", "neuron_9k"],
    "3.0.0": [
        "pbmc_1k_protein_v3",
        "pbmc_10k_protein_v3",
        "malt_10k_protein_v3",
        "pbmc_1k_v2",
        "pbmc_1k_v3",
        "pbmc_10k_v3",
        "hgmm_1k_v2",
        "hgmm_1k_v3",
        "hgmm_5k_v3",
        "hgmm_10k_v3",
        "neuron_1k_v2",
        "neuron_1k_v3",
        "neuron_10k_v3",
        "heart_1k_v2",
        "heart_1k_v3",
        "heart_10k_v3",
    ],
    "3.1.0": ["5k_pbmc_protein_v3", "5k_pbmc_protein_v3_nextgem"],
}

dataset_to_group = dict([(dataset_name, group)
                         for group, list_datasets in available_datasets.items()
                         for dataset_name in list_datasets])

group_to_url_skeleton = {
    "1.1.0":
        "http://cf.10xgenomics.com/samples/cell-exp/{}/{}/{}_{}_gene_bc_matrices.tar.gz",
    "2.1.0":
        "http://cf.10xgenomics.com/samples/cell-exp/{}/{}/{}_{}_gene_bc_matrices.tar.gz",
    "3.0.0":
        "http://cf.10xgenomics.com/samples/cell-exp/{}/{}/{}_{}_feature_bc_matrix.tar.gz",
    "3.1.0":
        "http://cf.10xgenomics.com/samples/cell-exp/{}/{}/{}_{}_feature_bc_matrix.tar.gz",
}

available_specification = ["filtered", "raw"]


def read_dataset10x_cellexp(name, spec, override=False, verbose=False):
  assert spec in available_specification, "Unknown specification '%s'" % spec
  if name == 'cellvdj':
    url = "http://cf.10xgenomics.com/samples/cell-vdj/3.0.2/" +\
          "vdj_v1_hs_aggregated_donor1/" + \
            "vdj_v1_hs_aggregated_donor1_%s_feature_bc_matrix.tar.gz"
    url = url % spec
  else:
    group = dataset_to_group[name]
    url_skeleton = group_to_url_skeleton[group]
    url = url_skeleton.format(group, name, name, spec)

  filename = os.path.basename(url)

  download_path = os.path.join(DOWNLOAD_DIR, name)
  if not os.path.exists(download_path):
    os.mkdir(download_path)

  preprocessed_path = os.path.join(PREPROCESSED_BASE_DIR,
                                   '%s_%s_preprocessed' % (name, spec))

  if override and os.path.exists(preprocessed_path):
    if verbose:
      print("Overriding path: %s" % preprocessed_path)
    shutil.rmtree(preprocessed_path)
  if not os.path.exists(preprocessed_path):
    os.mkdir(preprocessed_path)

  # ******************** preprocessed ******************** #
  if not os.path.exists(os.path.join(preprocessed_path, 'X')):
    path = get_file(fname=filename,
                    origin=url,
                    outdir=download_path,
                    verbose=bool(verbose))
    if not tarfile.is_tarfile(path):
      raise RuntimeError("Expecting tarfile but received: %s" % path)

    contents = {}
    with tarfile.open(path, mode="r:gz") as f:
      for info in f:
        if info.isfile():
          name = info.name
          if verbose:
            print("Extracting '%s' size %.2f(MB) ..." %
                  (name, info.size / 1024. / 1024.))

          data = f.extractfile(name)
          if is_gzip_file(data):
            data = gzip.open(data, mode="rb")
          name = os.path.basename(name).split('.')[0]

          if name == 'barcodes':
            data = np.array([str(line, 'utf-8')[:-1] for line in data])
          elif name == 'features' or name == 'genes':
            _ = []
            for line in data:
              line = str(line, 'utf-8')[:-1].split('\t')
              _.append(line)
            data = np.array(_)
          # this will take some time, better print out something
          elif name == 'matrix':
            data = mmread(data)
          else:
            raise RuntimeError(
                "Unknown downloaded file '%s', something changed from 10xGenomics."
                % name)
          contents[name] = data

    # cell barcodes
    X_row = contents['barcodes']
    # feature (Id, Name, Type(antibody or gene-expression))
    X_col = contents['features'] if 'features' in contents else contents['genes']
    # data matrix
    X = contents['matrix'].T
    if not isinstance(X, csr_matrix) and hasattr(X, 'tocsr'):
      X = X.tocsr()
    X = X.astype('float32')
    assert X.shape[0] == X_row.shape[0] and X.shape[1] == X_col.shape[0]

    prot_ids = []
    gene_ids = []
    # antibody and gene are provided
    if X_col.shape[1] == 3:
      for idx, row in enumerate(X_col):
        if row[-1] == 'Antibody Capture':
          prot_ids.append(idx)
        elif row[-1] == 'Gene Expression':
          gene_ids.append(idx)
        else:
          raise ValueError("Unknown feature type: %s" % str(row))
    else:
      gene_ids = slice(None, None)

    y = X[:, prot_ids]
    # Antibody ID, Antibody Name
    y_col = X_col[prot_ids][:, 0]
    y_col_name = X_col[prot_ids][:, 1]

    X = X[:, gene_ids]
    # Gene ID, Gene Name
    X_col_name = X_col[gene_ids][:, 1]
    X_col = X_col[gene_ids][:, 0]

    # smaller than 12000 samples, just convert to dense matrix
    # for fast processing
    if len(X_row) < 60000:
      if verbose:
        print(
            "Less than 60000 samples, convert all sparse matrices to dense ...")
      if issparse(X):
        X = X.todense()
      if issparse(y):
        y = y.todense()

    save_to_dataset(preprocessed_path,
                    X=X,
                    X_col=X_col,
                    y=y,
                    y_col=y_col,
                    rowname=X_row,
                    print_log=verbose)
    with open(os.path.join(preprocessed_path, 'X_col_name'), 'wb') as f:
      pickle.dump(X_col_name, f)
    if len(y_col_name) > 0:
      with open(os.path.join(preprocessed_path, 'y_col_name'), 'wb') as f:
        pickle.dump(y_col_name, f)
  # ******************** load and return the dataset ******************** #
  ds = Dataset(preprocessed_path, read_only=True)
  return ds



import base64
import os
import pickle
import shutil
import zipfile
from io import BytesIO

import numpy as np

from odin.fuel import Dataset
from odin.utils import batching, ctext, get_file, select_path
from odin.utils.crypto import decrypt_aes, md5_checksum
from sisua.data.path import DOWNLOAD_DIR, PREPROCESSED_BASE_DIR
from sisua.data.utils import remove_allzeros_columns, save_to_dataset

# ===========================================================================
# Const
# ===========================================================================
# top 5000 variable genes
_URL_5000 = b'aHR0cHM6Ly9zMy5hbWF6b25hd3MuY29tL2FpLWRhdGFzZXRzL0dTRTEwMDg2Nl9QQk1DLnJhd0Nv\ndW50RGF0YS41MDAwLmh2Zy5jc3Yuemlw\n'
_MD5_5000 = '46150f63e5a3c81d4f07445a759faa2b'

# raw Count Gene
_URL_FULL = b'aHR0cHM6Ly9zMy5hbWF6b25hd3MuY29tL2FpLWRhdGFzZXRzL0dTRTEwMDg2Nl9QQk1DLnJhd0Nv\ndW50RGF0YS5jc3Yuemlw\n'
_MD5_FULL = '7481cc9d20adef4d06fdb601d9d99e77'

# protein
_URL_PROTEIN = b'aHR0cHM6Ly9zMy5hbWF6b25hd3MuY29tL2FpLWRhdGFzZXRzL0dTRTEwMDg2Nl9QQk1DLnJhd0Nv\ndW50UHJvdGVpbi5jc3Yuemlw\n'
_MD5_PROTEIN = '7dc5f64c2916d864568f1b739679717e'

_CITEseq_PBMC_PREPROCESSED = select_path(os.path.join(
    PREPROCESSED_BASE_DIR, 'PBMC_citeseq_preprocessed'),
                                         create_new=True)
_5000_PBMC_PREPROCESSED = select_path(os.path.join(
    PREPROCESSED_BASE_DIR, 'PBMC_citeseq_5000_preprocessed'),
                                      create_new=True)

_PASSWORD = 'uef-czi'


# ===========================================================================
# Main
# ===========================================================================
def read_CITEseq_PBMC(override=False, verbose=False, version_5000genes=False):
  download_path = os.path.join(
      DOWNLOAD_DIR,
      "PBMC_%s_original" % ('5000' if version_5000genes else 'CITEseq'))
  if not os.path.exists(download_path):
    os.mkdir(download_path)

  preprocessed_path = (_5000_PBMC_PREPROCESSED
                       if version_5000genes else _CITEseq_PBMC_PREPROCESSED)
  if override:
    shutil.rmtree(preprocessed_path)
    os.mkdir(preprocessed_path)
  # ******************** preprocessed data NOT found ******************** #
  if not os.path.exists(os.path.join(preprocessed_path, 'X')):
    X, X_row, X_col = [], None, None
    y, y_row, y_col = [], None, None
    # ====== download the data ====== #
    download_files = {}
    for url, md5 in zip(
        [_URL_5000 if version_5000genes else _URL_FULL, _URL_PROTEIN],
        [_MD5_5000 if version_5000genes else _MD5_FULL, _MD5_PROTEIN]):
      url = str(base64.decodebytes(url), 'utf-8')
      base_name = os.path.basename(url)
      get_file(fname=base_name,
               origin=url,
               outdir=download_path,
               verbose=verbose)
      download_files[base_name] = (os.path.join(download_path, base_name), md5)
    # ====== extract the data ====== #
    n = set()
    for name, (path, md5) in sorted(download_files.items()):
      if verbose:
        print("Extracting %s ..." % ctext(name, 'lightcyan'))
      binary_data = decrypt_aes(path, password=_PASSWORD)
      md5_ = md5_checksum(binary_data)
      assert md5_ == md5, "MD5 checksum mismatch for file: %s" % name
      with zipfile.ZipFile(file=BytesIO(binary_data), mode='r') as f:
        for name in f.namelist():
          data = str(f.read(name), 'utf8')
          for line in data.split('\n'):
            if len(line) == 0:
              continue
            line = line.strip().split(',')
            n.add(len(line))
            if 'Protein' in name:
              y.append(line)
            else:
              X.append(line)
    # ====== post-processing ====== #
    assert len(n) == 1, \
    "Number of samples inconsistent between raw count and protein count"
    if verbose:
      print("Processing %s ..." % ctext("gene count", 'cyan'))
    X = np.array(X).T
    X_row, X_col = X[1:, 0], X[0, 1:]
    X = X[1:, 1:].astype('float32')

    # ====== filter mouse genes ====== #
    human_cols = [True if "HUMAN_" in i else False for i in X_col]
    if verbose:
      print("Removing %s MOUSE genes ..." %
            ctext(np.sum(np.logical_not(human_cols)), 'cyan'))
    X = X[:, human_cols]
    X_col = np.array([i for i in X_col if "HUMAN_" in i])
    X, X_col = remove_allzeros_columns(matrix=X,
                                       colname=X_col,
                                       print_log=verbose)

    # ====== protein ====== #
    if verbose:
      print("Processing %s ..." % ctext("protein count", 'cyan'))
    y = np.array(y).T
    y_row, y_col = y[1:, 0], y[0, 1:]
    y = y[1:, 1:].astype('float32')

    assert np.all(X_row == y_row), \
    "Cell order mismatch between gene count and protein count"

    # save data
    if verbose:
      print("Saving data to %s ..." % ctext(preprocessed_path, 'cyan'))
    save_to_dataset(preprocessed_path,
                    X,
                    X_col,
                    y,
                    y_col,
                    rowname=X_row,
                    print_log=verbose)
  # ====== read preprocessed data ====== #
  ds = Dataset(preprocessed_path, read_only=True)
  return ds
